"""This module handles loading, updating, and getting PyETC form defaults and options.
"""

import os, re, sys
from pandeia.engine.helpers.bit.pyetc_util import log, read_dict, write_dict

# ==========================================================================

_HERE = os.path.dirname(__file__) or "."

# ==========================================================================

class UnknownDefaults(Exception):
    pass

# ==========================================================================

class DictionaryMap:
    """Loads all the dictionaries associated with a file which lists one
    dictionary file per line.   Each line is of the form:
    <instrument> <sciencemode>  <dictionary_filename>
    """
    def __init__(self, kind, mapfile):
        self._kind = kind
        self._dict = {}
        self._map_paths = {}
        self._mapfile = mapfile
        for line in open(self._mapfile).readlines():
            instr, sci, fname = line.split()
            path = self._map_paths[(instr, sci)] = _HERE + "/" + fname
            try:
                self._dict[(instr,sci)] = read_dict(path)
            except:  # for generating new defaults for new modes
                log.warning("Error loading", self._kind,"for", (instr, sci))
                sys.exc_clear()

    def update_dict(self, instrument, sciencemode, d, hdr=[], outdir=None):
        """Updates the in-memory and file copy of a dict."""
        self._dict[(instrument, sciencemode)] = d
        path = self._map_paths[(instrument, sciencemode)]
        if outdir: # new output location?
            path = outdir+os.sep+os.path.basename(path)
        write_dict(self._force_strings(d), path, hdr)
        return path

    def _force_strings(self, d):
        if isinstance(d, list):
            return [ self._force_strings(i) for i in d]
        elif isinstance(d, dict):
            return dict([ (key,self._force_strings(d[key])) for key in d])
        else:
            return str(d)

    def get_dict(self, instrument, sciencemode, **keys):
        """Return the default test input parameters for this test class."""
        try:
            defs = self._dict[(instrument, sciencemode)]
        except KeyError:
            raise UnknownDefaults("Unknown " + self._kind + " for ", (instrument,sciencemode))
        return dict(defs.items())

# ==========================================================================

_DEFAULTS = DictionaryMap("defaults", _HERE + "/map")
_OPTIONS = DictionaryMap("options", _HERE + "/options_map")

update_defaults = _DEFAULTS.update_dict
get_defaults = _DEFAULTS.get_dict

update_options = _OPTIONS.update_dict
get_options = _OPTIONS.get_dict



import pandeia.engine.helpers.bit.instruments

#
# read a dictionary from one of the .dat files

def get_data( name ) :
    return pandeia.engine.helpers.bit.instruments.get_instrument_data( 'acs', name )

FILTERS = get_data("config")["filters"]
DETECTOR_TYPE = dict(hrc='ccd',wfc='ccd',sbc='mama')

def instrument_from_form(science_mode, form_input):
    """Function to extract values of interest from the web form
    data and return a smaller dictionary which will be used to
    update the engine_input dictionary.
    """
    idict = dict( gain =   form_input['gain'],
                  nreads = form_input['crsplit'],
                  fdiameter = form_input['fdiameter'],
                  )
    idict["wodict"] = wodict = {}
    wodict["coronography"] = "off"

    #check if nreads is zero and return an error
    if idict["nreads"]  < 1:
        raise ValueError("CRSPLIT cannot be less than 1")

    if science_mode == 'imaging':
        #make an obsmode string that can be submitted to pysynphot
        idict['obsmode'], filter, polarizer = make_acs_imaging_obsmode(form_input)
        idict['detector'] = form_input['detector']
        detector = idict['detector']
        if idict['detector'].lower() == 'sbc':
            idict['gain']=1.
        idict["filter"] = wodict["filter"] = filter
        if filter is not None:
            idict["filter"] = filter.lower()
        elif polarizer is not None:
            # this should be a case of clear filter plus polarizer so use
            # polarizer for bandpass limits.
            idict["filter"] = polarizer.lower()
        else:
            # this should be the case where there are two clear filters so pick one.
            if detector == 'wfc':
                idict['filter']=form_input['wfcfilt1'].lower()
            elif detector == 'hrc':
                idict['filter']=form_input['hrcfilt1'].lower()
        wodict["filter"] = idict["filter"]
        if filter is not None:
            wodict["filter_info"] = FILTERS[filter]
    elif science_mode == 'rampfilter':
        #make an obsmode string that can be submitted to pysynphot
        idict['obsmode']=make_acs_rampfilter_obsmode(form_input)
        wodict["rampfilter"] = True
        detector=form_input['detector']
        idict['detector'] = detector
        if detector == 'wfc':
            idict['filter']=form_input['wfcfilt1'].lower()
        elif detector == 'hrc':
            idict['filter']=form_input['hrcfilt1'].lower()
        wodict["filter"] = idict["filter"]
        wodict["filter_info"] = FILTERS[idict["filter"]]
        polarizer = None
    elif science_mode == 'spectroscopic':
        #the spectroscopic form is set up so that the user chooses a disperser
        #and the name of the disperser contains the name of the detector. I don't
        #like this, but am keeping it for now.
        combine_string=(form_input['disperser']).split('_')
        detector=combine_string[0].lower()
        grating=combine_string[1].lower()[1:]
        idict['obswave']=form_input['obswave']
        idict['grating']=grating
        idict['detector']=detector

        if detector == 'wfc':
            filter=form_input['wfcfilt1'].lower()
        elif detector == 'hrc':
            selector = {"0":"1", "1":"0"}[combine_string[1][0]]
            filter=form_input['hrcfilt' + selector].lower()
            wodict["filter"] = filter
        elif detector == 'sbc':
            filter='clear' #to pass the concatination
            idict['gain']=1
        idict["filter"] = filter
        #make an obsmode string that can be submitted to pysynphot
        idict['obsmode'] = make_acs_spectroscopic_obsmode(idict)

        wodict["disperser_info"] = disperser = {}
        disperser["label"] = {"p" : "Prism", "g" : "Grism"}[grating[0]]
        disperser["grating"] = grating.upper()
        disperser["name"] = FILTERS[grating]["name"]
        if not filter.startswith("clear"):
            polarizer = filter
        else:
            polarizer = None
    else:
        raise NotImplementedError('%s mode not implemented for acs' % science_mode)
    if polarizer:
        wodict["polarizer"] = {
            "pol_v":"Visible",
            "pol_uv":"Ultraviolet"
        }[polarizer]

    wodict['detector_type'] = DETECTOR_TYPE[idict['detector']]

    # Silently discard post_flash inputs for non-CCDs and non-supported spectrograph mode
    # See #922
    wodict["show_post_flash"] = False
    # NOTE that w/ show_post_flash, we add a control-flow flag in a place
    # where its not expected. During code review jwcalibdev.stsci.edu/r/69
    # we decided to just comment for now.  If it happens again, we may want
    # to add a dict under/near wodict to handle these things.
    if science_mode != 'spectroscopic' and \
       DETECTOR_TYPE[idict['detector']] == 'ccd':
        idict['post_flash_electrons'] = form_input['post_flash_acs']
        wodict["show_post_flash"] = True

    return idict

def make_acs_imaging_obsmode(form_input):
    # instrument
    elements = ['acs']
    #The detector gets an entry of its own on only some of the forms
    #on other forms it derived from the value of one of the other
    #varibles (as in spec  mode). If that's the case pass in the detector
    #to this function.
    detector = form_input['detector']
    #QUERY: wfc1 and wfc2 are distinct obsmodes,
    #representing the two halves of the chip. But the
    #ETC always assumes wfc1.
    if detector == 'wfc':
        elements.append('wfc1')
    else:
        elements.append(detector)
    #filters: Based on the detector, we decide which filters to pick out.
    f0 = form_input['%sfilt0' % detector].lower()
    if detector != 'sbc':
        #there are two filter wheels, either of which may be clear
        f1 = form_input['%sfilt1' % detector].lower()
        filters = [f for f in (f0, f1) if not f.startswith('clear')]
        elements.extend(filters)
    else:
        #there is only one and it can't be clear
        elements.append(f0)
        filters = [f0]
    obsmode = ','.join(elements)

    # For now anyway,  polarizers appear on the form with a phase but are
    # only uploaded to the server as V or UV.
    polarizer = filter = None
    filter_count = 0
    for f in filters:
        if f.startswith("pol"):
            polarizer = f
        elif not f.startswith("clear"):
            filter = f
            filter_count += 1

    #Enforce restriction against crossed filters
    if filter_count > 1:
        raise ValueError("Use of two non-clear filters %s is not supported. Please select only one filter and an optional polarizer." % str(filters))

    return obsmode, filter, polarizer

def make_acs_rampfilter_obsmode(form_input):
    #we need a string like: acs,wfc,fr388n#3880

    #instrument
    elements = ['acs']
    #Detector
    detector = form_input['detector']
    #wfc always uses wfc1
    if detector == 'wfc':
        elements.append('wfc1')
    else:
        elements.append(detector)

    #filter: choose based on detector
    f = form_input['%sfilt1'%detector].lower()
    #and the wavelength for the ramp filter
    wave=form_input['obswave']

    elements.append("%s#%s"%(f,wave))
    obsmode = ','.join(elements)

    # There is a KNOWN BUG in pysynphot right now that
    # does not support the full range of wavelengths for the
    # ramp filters. So try to construct a bandpass now, so
    # we can raise an appropriate exception if it fails.
    #
    # This is problematic because it requires importing
    # pysynphot into this web module, which in general is
    # disallowed for clean-interface reasons. This code
    # should be removed as soon as the known pysynphot bug
    # is fixed.
    #
    # TODO: remove this code when pysynphot is fixed
    # TODO: think about better error message management.
    import pysynphot as psyn
    try:
        psyn.ObsBandpass(obsmode)
    except psyn.exceptions.ExtrapolationNotAllowed as e:
        msg = """You have selected instrument configuration %s at a wavelength of %g. This wavelength is outside the range
              of this filter. Please select a wavelength that is within the supported range for this filter."""
        raise ValueError(msg%(str(elements[:]), wave))
    return obsmode

def make_acs_spectroscopic_obsmode(form_input):
    #instrument
    elements = ['acs']

    # This function can be called with a 'processed' input dictionary,
    # as well as a raw one. The raw dictionary contains exactly what
    # is in the input page. The input spectroscopic page has no 'detector'
    # 'grating', or 'filter' fields, so we get them from the 'disperser'
    # string. This is similar to what is done in instrument_from_form()
    # above, but not quite the same. An effort should be put here when
    # refactoring the web code.
    if 'detector' in form_input:
        detector = form_input['detector']
        grating = form_input['grating']
        filter = form_input['filter']
    else:
        combine_string = (form_input['disperser']).split('_')
        detector = combine_string[0].lower()
        grating = combine_string[1].lower()[1:]
        if detector == 'wfc':
            filter = form_input['wfcfilt1'].lower()
        elif detector == 'hrc':
            selector = {"0":"1", "1":"0"}[combine_string[1][0]]
            filter = form_input['hrcfilt' + selector].lower()
        elif detector == 'sbc':
            filter = 'clear'
        else:
            filter = ''

    #detector: wfc always uses wfc1
    if detector == 'wfc':
        elements.append('wfc1')
    else:
        elements.append(detector)

    #Based on the detector, we decide which filters to pick out.
    f = [x for x in (grating, filter) if not x.startswith("clear")]
    elements.extend(f)
    obsmode = ','.join(elements)
    return obsmode



#ACS limits.

# This file is read when the instruments/data.py file is executed
# as part of the instrument module import. The resulting dictionary
# is used to initialize instances of Limit that will be used to
# check any subsequent ACS calculation.

# allsub defines what the magic value "all" should expand into
allsub = {'mode': ['imaging', 'rampfilter', 'spectroscopic'],
          'detector': ['hrc', 'wfc', 'sbc']
}

# limit_spec contains the actual limit definitions. It is a
# dictionary {limit shortname : {definition of limit} }.
# See user's guide for detailed requirements of the limit definition.

# Message strings may use the following variable names for substitution:
#  varname = value of the variable specified in "varname"
#  threshold = value specified in "threshold"
#  percentval = percentage expression of (1-subthresh)
#
limit_spec = {
    'mintime': {
        'varname': 'time',
        'message': '''Result of the calculation %(varname)g seconds per frame is less than the minimum
                  exposure time for this detector (%(threshold)g seconds) ''',
        'mincheck': True,
        'mode': 'all',
        'varieties': (
                {'detector': 'hrc',
                 'per_read': True,
                 'threshold': 0.1, },

                {'detector': 'wfc',
                 'per_read': True,
                 'threshold': 0.5}
                 )
    }, #end mintime

    'maxtime': {
        'varname': 'time',
        'message': '''Result of the calculation %(varname)g seconds is more than the maximum
                  exposure time for this detector (%(threshold)g seconds) ''',
        'mode': 'all',
        'varieties': (
                {'detector': 'hrc',
                 'per_read': True,
                 'threshold': 3600., },
                {'detector': 'wfc',
                 'per_read': True,
                 'threshold': 3600.}
                    )
    },
    # end maxtime

    'brightpix': {
        'varname': 'brightest_pixel_counts',
        'mode': 'all',
        'message': '''CCD saturation limit: total electrons per pixel exceeds
                full-well limit: value=%(varname)g, full-well limit=%(threshold)g''',
        'subthresh': 0.7,
        'submessage': '''This exposure exceeds %(percentval)3g percent of the CCD
             full-well limit: value=%(varname)g, limit=%(threshold)g''',
        'varieties': ( #hrc thresholds
                       {
                           'detector': 'hrc',
                           'threshold': ('gain', {
                               1: 65535,
                               2: 131000,
                               4: 155000,
                               8: 155000} )
                       },
                       # wfc thresholds
                           {'detector': 'wfc',
                            'threshold': ('gain', {
                                0.5: 32700,
                                1.0: 65535,
                                1.4: 84700,
                                2.0: 84700, } )
                       },
                       # sbc thresholds
                           {'detector': 'sbc',
                            'varname': 'brightest_pixel_rate',
                            'threshold': 50,
                            'message': '''MAMA Bright limits: total count rate %(varname)g per binned pixel
                       exceeds bright limit %(threshold)g''',
                            'subthresh': 0.44, #=value of 22
                            'submessage': '''MAMA Bright limits: total count rate %(varname)g per binned pixel exceeds %(percentval)d percent
                         of the bright limit %(threshold)g. The detector is known to be non-linear in this domain,
                         and reliable photometry cannot be extracted.''',

                            }, # end sbc
            ) # end varieties
    }, # end brightpix

    'brightimg': {
        'varname': 'detector_total_rate',
        'detector': 'sbc',
        'mode': 'all',
        'threshold': 200000,
        'message': "MAMA Bright limits: total count rate per image %(varname)g exceeds bright limit %(threshold)g",
        'subthresh': 0.6,
        'submessage': "MAMA Bright limits: total count rate per image %(varname)g exceeds %(percentval)d percent of bright limit %(threshold)g and would exceed it for irregular variable sources"
        ,
        }, #end brightimg

    'waverange': {
        'varname': 'effective_wavelength',
        'detector': 'sbc',
        'mode': 'all',
        'threshold': 3000,
        'message': "Please note that the effective wavelength of this simulation %(varname)g is beyond %(threshold)g Angstroms. The SBC PSF has not been characterized in this spectral range and therefore large uncertainties are possible in the calculation of the S/N for this observation."
    }, #end waverange

    'minbackground': {
        'varname': 'background_pixel_counts',
        'mincheck': True,
        'message': "Electrons per pixel due to background (%(varname).2g) is less than the recommended threshold \
        of %(threshold)g electrons to avoid poor charge transfer efficiency (CTE). In rare cases, post-flash may \
        be recommended.  Please see the Post-Flash Recommendations section of ACS ISR 2014-01, which can be found on the ACS website.",
        'mode': 'all',
        'per_read': True,
        'varieties': (
            {'detector': 'hrc',
             'threshold': 20.0
        },
            {'detector': 'wfc',
             'threshold': 20.0
            }
        )
    }, #end minbackground

    'maxleak' : {
        'varname' : 'total_leak',
        'message' : '''This observation has a significant filter leak where
                       %(varname).3g percent of the signal comes from outside the prescribed
                       filter bandpass.''',
        'mode' : 'imaging',
        'detector' : 'all',
        'per_read' : False,
        'threshold' : 5.0,
        }, # end maxleak

    'zerotargetflux' : {
        'varname' : 'target_total_rate',
        'mincheck' : True,
        'mode' : 'imaging',
        'detector' : 'all',
        'per_read' : False,
        'threshold' : 1.E-25,
        'varieties': (
            {'mode'   : 'imaging',
             'message': '''Zero target flux in the observation bandpass''',},
            {'mode'   : 'spectroscopic',
             'message': '''Zero target flux at the observation wavelength''',},
        ),
       }, # end zerotargetflux

    # this is not a real limit, but a warning associated with the filter/detector/science mode
    # combination. We fake it as a zero-threshold limit so the existing infrastructure
    # for limit processing can be re-used. This infrastructure creates as main product
    # a warning message with the desired properties.
    'clearfilters' : {
       'varname' : 'target_total_rate',
       'mode'    : ['imaging','rampfilter'],
       'filter'  : ['clear2l','clear2s'],
       'varieties' : (
           {
             'detector': 'hrc',
             'threshold': 0,},
           {
             'detector': 'wfc',
             'threshold': 0,},
           ),
       'message' : '''
            Filterless operation of any ACS channel (CLEAR selected for both filter wheels) is 
            not supported, and will result in a severely degraded point spread function (PSF). 
            See Section 7.1.2 of the ACS Instrument Handbook and ACS ISR 2003-03 for more details.
       ''', 
    }, #end clearfilters 

} # end limit_spec



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from functools import reduce

import astropy
import numpy as np
import synphot as syn
from synphot.models import Empirical1D, ConstFlux1D
from .constants import PANDEIA_WAVEUNITS, PANDEIA_FLUXUNITS

from .custom_exceptions import EngineInputError, SynphotError

default_separator = "__"


def spectrum_resample(flux, orig_wave, new_wave, mask_val=np.nan):
    """
    Use synphot to re-sample a spectrum to a new set of wavelengths while conserving flux.
    Parameters
    ----------
    flux: 1D np.ndarray
        Input spectrum to be re-binned
    orig_wave: 1D np.ndarray
        Set of wavelengths for input spectrum
    new_wave: 1D np.ndarray
        New set of wavelengths to re-bin spectrum onto
    mask_val: float or np.nan (default: np.nan)
        Value to fill in where new_wave is outside the bounds of orig_wave. np.nan is the right thing
        to use here, but make it configurable in case it needs to be changed.
    Returns
    -------
    binned_flux: 1D np.ndarray
        Input spectrum re-binned onto new_wave
    """
    # if wavelength sets are the same, then pass back input flux unmodified.  the input wavelengths set the nyquist limit
    # so if you resample a spectrum back onto it's own wavelength set, it still gets downgraded if it was undersampled.
    if (orig_wave.size == new_wave.size) and np.allclose(orig_wave, new_wave):
        binned_flux = flux
    else:
        try:
            # synphot will erroneously assume the data is in Angstroms and PHOTLAM, but this will enable it to 
            # actually rebin the data correctly. Problems (NaNs) arise when it tries to convert mJy (an fnu unit) 
            # to PHOTLAM (an flam unit), binning, and converting back to mJy. (JETC-300)
            # See similar code in helpers/background/butler.py bg_resample.
            spec = syn.spectrum.SourceSpectrum(Empirical1D, points=orig_wave, lookup_table=flux)
            dummy_filter = syn.spectrum.SpectralElement(Empirical1D, points=orig_wave, lookup_table=np.ones_like(flux))
            # We can't just resample the flux, because that performs linear interpolation (Empirical1D model is based on
            # the astropy Tabular1D model, which according to this page:
            # https://docs.astropy.org/en/stable/modeling/reference_api.html#astropy.modeling.tabular.Tabular1D just does
            # linear interpolation. The Observation with provided binset does flux-conserving rebinning
            obs = syn.observation.Observation(spec, dummy_filter, binset=new_wave, force="extrap")
            # obs(new_wave) would just resample the spectrum within the observation. We specifically need obs.binflux
            # Again, this is actually in mJy, but synphot thinks it's PHOTLAM.
            binned_flux = obs.binflux.value
        except Exception as e:
            msg = "Error resampling spectrum to new waveset via Synphot: %s" % type(e)
            raise SynphotError(value=msg)

        # the use of 'extrap' means synphot will fill in the last available flux value for any wavelengths beyond the 
        # bounds of the original spectrum (its alternative is raising an error). This is going to be wrong for just 
        # about any circumstance. The right thing to do is to recognize that we don't know what's beyond the bounds 
        # of the original spectrum and use np.nan as the fill value for fluxes at these new wavelengths. mask_val 
        # defaults to np.nan, but is configurable in case there's a need to use something else.
        wmin = orig_wave.min()
        wmax = orig_wave.max()
        invalid_subs = np.where((new_wave < wmin) | (new_wave > wmax))
        binned_flux[invalid_subs] = mask_val
    return binned_flux


def recursive_subclasses(cls):
    """
    The __subclasses__() method only goes on level deep, but various classes that ultimately
    inherit from things like Instrument and Strategy are separated by multiple inheritance layers.
    This function recursively walks through the inheritance tree and returns a list of all subclasses
    at all levels that inherit from the given class.

    Parameters
    ----------
    cls: any python Class
        Python class to get list of subclasses for

    Returns
    -------
    all_subclasses: list
        List of all subclasses that ultimately inherit from cls
    """
    all_subclasses = []

    top_subclasses = cls.__subclasses__()
    all_subclasses.extend(top_subclasses)

    for s in top_subclasses:
        all_subclasses.extend(recursive_subclasses(s))

    return all_subclasses


def merge_data(*dicts):
    """
    This takes a list of python dicts and merges them into a single dict.  It is set up to
    assure that later arguments will take precedence over earlier ones by default.

    Parameters
    ----------
    dicts: list
        List of dicts to merge into a single dict

    Returns
    -------
    updated: dict
        Arguments combined into a single dict.  Later arguments take precedence over earlier arguments
        and dicts take precedence over non-dict values.
    """
    updated = {}

    # grab all of the keys
    keys = set()
    for o in dicts:
        keys = keys.union(set(o))

    for key in keys:
        values = [o[key] for o in dicts if key in o]
        # find values that are dicts so we can recurse through them
        maps = [value for value in values if isinstance(value, dict)]
        if maps:
            updated[key] = merge_data(*maps)
        else:
            # if not a dict, then return the last value we have since later arguments
            # take precedence
            updated[key] = values[-1]
    return updated


def apply_default_only(input_dict):
    """
    Recursively apply the special default_only parameters

    This is a special section for handling locations where there are two different keywords that can't
    coexist, like specifying scan_length XOR scan_rate in a scan calculation. The pandeia.engine Instrument
    default loader can't add both, and doesn't know which one the user might choose, so it must require
    neither. But at least one MUST be present, so we need to specify one to add when we build a calculation
    (here).

    Paramters
    ---------
    input_dict: dictionary
        A dictionary from somewhere in the engine input dictionary

    Returns
    -------
    input_dictionary: dictionary
        The same dictionary with "default_only" values moved into the main dictionar(ies), and the default_only
        subdict removed.

    """
    if "default_only" in input_dict:
        input_dict.update(input_dict['default_only'])
        # and now delete the extraneous dictionary
        del input_dict['default_only']
    for item in input_dict:
        if isinstance(input_dict[item], dict):
            input_dict[item] = apply_default_only(input_dict[item])
    return input_dict


def get_key_list(key, separator=default_separator):
    """
    Create list of keys from a flattened key to set/retrieve value from nested dict
    """
    keylist = key.split(separator)
    return keylist


def flat_key_from_list(list, separator=default_separator):
    """
    Create a flattened version of a nested key from a key list
    """
    flat_key = separator.join(list)
    return flat_key


def get_dict_from_keys(data, keylist):
    """
    Use a list of keys to recurse through a dict to get a value

    >>> d = {'a': {'b': 2}}
    >>> get_dict_from_keys(d, ['a', 'b'])
    2
    """
    try:
        val = reduce(dict.__getitem__, keylist, data)
    except KeyError as e:
        msg = "Invalid keys provided: %s (%s)" % (repr(keylist), e)
        raise EngineInputError(value=msg)
    return val


def set_dict_from_keys(data, keylist, value):
    """
    Use a list of keys to recurse through a dict to set a value

    >>> d = {'a': {'b': 2}}
    >>> set_dict_from_keys(d, ['a', 'b'], 6)
    >>> print(d)
    {'a': {'b': 6}}
    """
    get_dict_from_keys(data, keylist[:-1])[keylist[-1]] = value


def lower_key(in_obj):
    """
    take input object and recursively make keys and strings lower-case and convert
    spaces to underlines to ensure compliance with pyetc coding standards.

    Parameters
    ----------
    in_obj: python object
        object containing keys and values to be modified. if it is not a dict, list, str, or unicode
        it is simply returned to the caller unchanged.

    Returns
    -------
    obj: python object
        object of same type as in_obj with strings modifyed to be lower case and spaces removed
    """
    if isinstance(in_obj, dict):
        out_dict = {}
        for key, item in list(in_obj.items()):
            out_dict[key.lower().replace(' ', '_')] = lower_key(item)
        obj = out_dict
    elif isinstance(in_obj, list):
        obj = [lower_key(obj) for obj in in_obj]
    elif isinstance(in_obj, str) or isinstance(in_obj, str):
        obj = in_obj.lower().replace(' ', '_')
    else:
        obj = in_obj
    return obj



#wfc3uvis  limits

# This file is read when the instruments/data.py file is executed
# as part of the instrument module import. The resulting dictionary
# is used to initialize instances of Limit that will be used to
# check any subsequent WFC3UVIS calculation.

# allsub defines what the magic value "all" should expand into
allsub = {'mode': ['imaging', 'spectroscopic', 'scimaging', 'scspectroscopic'],
          'detector' : ['uvis1','uvis2']
         }

# limit_spec contains the actual limit definitions. It is a
# dictionary {limit shortname : {definition of limit} }.
# See user's guide for detailed requirements of the limit definition.

# Message strings may use the following variable names for substitution:
#  varname = value of the variable specified in "varname"
#  threshold = value specified in "threshold"
#  percentval = percentage expression of (1-subthresh)
#

limit_spec = {

  # A reference for the UVIS min/max times is also the Phase II
  # Proposal Instructions for Cycle 18, section 14.2.5, where it states
  # the allowed range of 0.5 to 3600 sec.
  'mintime' : {
      'varname' : 'time',
      'message' : '''Result of the calculation %(varname)g is less than the minimum
                  exposure time for this detector (%(threshold)g) ''',
      'mincheck' : True,
      'mode' : 'all',
      'detector' : 'all',
      'per_read' : True,
      'threshold': 0.5,
     }, # end mintime

  'maxtime' : {
      'varname' : 'time',
      'message' : '''Result of the calculation %(varname)g is more than the maximum
                  exposure time for this detector (%(threshold)g) ''',
      'mode' : 'all',
      'detector' : 'all',
      'per_read' : True,
      'threshold': 3600.,
     }, #end maxtime
  
    # Saturation values (brightpix) are taken from the WFC3 Instrument Handbook
    # for Cycle 19,section 5.4.5, which states that the lower-limit to the
    # spatially-varying saturation level across both CCD chips is 63,000e,
    # which is the value adopted for the ETC.
    'brightpix': {
    'varname' : 'brightest_pixel_counts',
    'message' : 'This observation has total electrons per pixel %(varname)g, which exceeds the saturation limit %(threshold)g.',
    'mode': 'all',
    'varieties' : ( #one for each detector
       { 'detector': 'uvis1',
            'threshold': 64000,
            'subthresh': 0.9,
            'submessage': 'This exposure (%(varname)g) exceeds %(percentval)d percent of the detector saturation limit (%(threshold)g)'
        }, 
       { 'detector': 'uvis2',
            'threshold': 67000,
            'subthresh': 0.9,
            'submessage': 'This exposure (%(varname)g) exceeds %(percentval)d percent of the detector saturation limit (%(threshold)g)'
        }
      ) #end varieties  
    }, #end brightpix

    'minbackground' : {
        'varname' : 'background_pixel_counts',
        'mincheck' : True,
        'message' : '''"Electrons per pixel due to background (%(varname).2g) is less than the recommended threshold
                        of %(threshold)g electrons to avoid poor charge transfer efficiency (CTE). We suggest you
                        consider CTE mitigation strategies described in Section 6.9 of the WFC3 Instrument Handbook.
                        Updates are provided on the WFC3 webpages."''',
        'mode' : 'all',
        'detector' : 'all',
        'per_read' : True,
        'threshold' : 20.0 
        }, #end minbackground

    'maxleak' : {
        'varname' : 'total_leak',
        'message' : '''This observation has a significant filter leak where
                       %(varname).3g percent of the signal comes from outside the prescribed
                       filter bandpass.''',
        'mode' : ['imaging', 'scimaging'],
        'detector' : 'all',
        'per_read' : False,
        'threshold' : 5.0,
        }, # end maxleak

    'zerotargetflux' : {
        'varname' : 'target_total_rate',
        'mincheck' : True,
        'mode' : ['imaging', 'scimaging'],
        'detector' : 'all',
        'per_read' : False,
        'threshold' : 1.E-25,
        'varieties': (
            {'mode'   : 'imaging',
             'message': '''Zero target flux in the observation bandpass''',},
            {'mode'   : 'spectroscopic',
             'message': '''Zero target flux at the observation wavelength''',},
        ),
       }, # end zerotargetflux

    'maxscanrate' : {
        'mode' : ['scimaging'],
        'detector' : 'all',
        'varname' : 'scan_rate',
        'threshold' : 7.84,
        'message' : '''Scan rate must be lower than %(threshold).3g arcsec/sec''',
        }, # end maxscanrate

    'gyroscanrate' : {
        'mode' : ['scimaging'],
        'detector' : 'all',
        'varname' : 'scan_rate',
        'threshold' : 5.,
        'message' : '''   'Scan rates larger than %(threshold).3g arcsec/sec are limited to gyro mode only''',
        }, # end gyroscanrate

 } #end limit_spec



import os
telescope = os.environ.get("ETC_TELESCOPE", "jwst").lower()

if telescope in ('jwst','roman'):
    from pandeia.engine.helpers.background.jwst.sl_accessor_funcs import get_dateless_bg
    from pandeia.engine.helpers.background.jwst.straylight_accessor import get_stray_light_bg
    from pandeia.engine.helpers.background.jwst.thermal_accessor import get_thermal_bg
    from pandeia.engine.helpers.background.jwst.bmg_accessor import get_in_field_bg
    from pandeia.engine.custom_exceptions import BMGError
    from pandeia.engine.custom_exceptions import BackgroundError
    from pandeia.engine.custom_exceptions import DatelessBGError
    from pandeia.engine.custom_exceptions import StraylightPositionError
    from pandeia.engine.custom_exceptions import StraylightDataError
    import synphot as syn
    import astropy.units as u
    from synphot.models import ConstFlux1D, Empirical1D
    from operator import add
elif telescope == 'hst':
    from pandeia.engine.helpers.background.hst.sky import Sky
    from pandeia.engine.helpers.background.hst.sky_spectrum import SkySpectrum


import numpy as np


'''
We have to resample the three background components because the infield waveset
is not the same as straylight and thermal.  So we normalize the waveset and then
resample each component.
'''
def bg_resample(merged, wave, flux):
    merged = np.array(merged)
    wave = np.array(wave)
    flux = np.array(flux)
    # we are NOT converting these spectra to microns and MJy (which should be MJy/sr anyway)
    # because that will trigger the internal unit conversion routines, and they have been found
    # to corrupt the spectra. Because we do not need to do any actual unit conversions, we can 
    # allow synphot to think these are its default units (Angstroms and PHOTLAM) and simply 
    # accept the rebinned values. (JETC-300) See similar code in engine/utils.py spectrum_resample
    spec = syn.spectrum.SourceSpectrum(Empirical1D, points=wave, lookup_table=flux)
    filt = syn.spectrum.SpectralElement(ConstFlux1D, amplitude=1)
    obs = syn.observation.Observation(spec, filt, binset=merged, force='taper')
    flux = obs.binflux.value
    return list(flux)


def call_butler(ra, dec, date, level, ra_dec_str, date_str):

    if date is not None:
        # dated

        # stray light
        try:
            sl_wave, sl_bg = get_stray_light_bg(ra, dec, date, ra_dec_str, date_str)
        except (StraylightPositionError, StraylightDataError):
            raise
        except Exception as e:
            raise BackgroundError('Error calculating stray light background.', e)

        # thermal
        try:
            thermal_wave, thermal_bg = get_thermal_bg()
        except Exception as e:
            raise BackgroundError('Error calculating thermal background.', e)

        # in field
        try:
            if_wave, if_bg = get_in_field_bg(ra, dec, date, ra_dec_str, date_str)
        except BMGError as e:
            raise BackgroundError('BMGError calculating infield background.  This can happen if the BMG is unavailable, or because %s.' % e)
        except Exception as e:
            if "timed out" in str(e):
                raise BackgroundError("Error calculating infield background: dated background is currently unavailable.  Try again later or switch to low/medium/high")
            else:
                raise BackgroundError('Error calculating infield background: %s' % e)

        # merge wavelengths
        # wave = syn.units.merge_wavelengths(thermal_wave, if_wave)
        wave = if_wave

    else:

        # get all components via special get_dateless_bg route if no date given
        try:
            unused_iday, wave, if_bg, sl_bg, thermal_wave, thermal_bg = get_dateless_bg(ra, dec, level)
            # the get_dateless_bg() API returns iday (int day of yr) which we don't currently use
        except (DatelessBGError, StraylightPositionError, StraylightDataError):
            raise
        except Exception as e:
            raise BackgroundError('Error calculating dateless background.', e)
        # wave stands for both if_wave (infield) and sl_wave (straylight)

    # resample the flux for each onto the new set
    if date is not None:
        sl_bg = bg_resample(wave, sl_wave, sl_bg)
    thermal_bg = bg_resample(wave, thermal_wave, thermal_bg)
    # if_bg = bg_resample(wave, if_wave, if_bg)

    # then sum the backgrounds
    combined_bg = list(map(add, thermal_bg, if_bg))
    combined_bg = list(map(add, combined_bg, sl_bg))

    # this will be written to a .npz file
    data_to_save = dict(
        straylight=sl_bg,
        thermal=thermal_bg,
        infield=if_bg,
        background=combined_bg,
        wavelength=wave
    )

    return [list(wave), combined_bg], data_to_save


def get_jwst_background(background):
    data_to_save = None

    # dated
    if background['bg_type'] == 'dated':
        ra = background['ra']
        dec = background['dec']
        ra_dec_str = background['ra_dec_str']
        date = background['date']
        date_str = background['date_str']
        background, data_to_save = call_butler(ra, dec, date, None, ra_dec_str, date_str)

    # dateless
    elif background['bg_type'] is not None and background['bg_type'].lower() != 'none':
        ra = background['ra']
        dec = background['dec']
        ra_dec_str = background['ra_dec_str']
        level = background['bg_type'].upper()[0]  # dateless code expects level of: L,M,H
        background, data_to_save = call_butler(ra, dec, None, level, ra_dec_str, "dateless")

    else:
        # Handle "none", but note that this is how we'd send positionless/dateless
        background = background['bg_type']

    return background, data_to_save


def get_hst_background(bg_input):
    data_to_save = None
    background = np.array([0.01, 30.0]), np.array([0.0, 0.0])

    bg = Sky(bg_input)
    bg.make_synexpr()

    if bg.expr is not None:
        bg_spectrum = SkySpectrum(bg.expr)
        # convert from angstroms and flam/arcsec^2 to microns and MJy/sr
        wavelength = bg_spectrum.wave_in_microns
        flux = bg_spectrum.flux_in_MJy_per_sr
        background = wavelength, flux

        # this will be written to a .fits file
        data_to_save = dict(
            background=flux,
            wavelength=wavelength
        )

    return background, data_to_save


def get_background(background):
    data_to_save = None

    if telescope == 'hst':
        background, data_to_save = get_hst_background(background)
    else:
        background, data_to_save = get_jwst_background(background)

    return background, data_to_save



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import copy

import numpy as np
from warnings import warn
import scipy.integrate as integrate
import scipy.interpolate as sci_int
from scipy.ndimage import shift
import scipy.signal as sg
from astropy.io import fits

from .config import DefaultConfig
from . import astro_spectrum as astro
from . import background as bg
from . import coords
from .custom_exceptions import EngineInputError, EngineOutputError, RangeError, DataError, DataConfigurationError
from .hst import HSTInstrument
from .pandeia_warnings import signal_warning_messages as warning_messages
from . import debug_utils




class CalculationConfig(DefaultConfig):
    """
    Encapsulate calculation configuration parameters (e.g. effects to include, noise
    sources to consider)
    """
    pass

class DetectorSignal(astro.ConvolvedSceneCube):
    """
    This class contains functionality for calculating the integrated electron rate of an
    astronomical source at the relevant instrument detector plane.

    Parameters
    ----------
    observation: observation.Observation instance
        Contains information required to configure a calculation
    calc_config: CalculationConfig instance
        Contains boolean flags that control which noise components are included in the
        calculation
    webapp: bool
        Toggle strict engine API checking
    validate: bool
        Keyword controlling whether intermediate products are saved for validation.
    order: int or None
        For multi-order case (i.e. SOSS), designate which order to use

    Attributes
    ----------
    """

    def __init__(self, observation, calc_config=CalculationConfig(), webapp=False, validate=False, order=None, empty_scene=False):
        self.validate = validate
        self.webapp = webapp

        # Get calculation configuration
        self.calculation_config = calc_config

        if self.validate:
            self.observation = observation

        # Load the instrument we're using
        self.current_instrument = observation.instrument
        # and configure it for the order we wish to use, if applicable
        self.current_instrument.order = order

        # save to the DetectorSignal instance, for convenience purposes
        self.order = order

        # how are we projecting the signal onto the detector plane?
        self.projection_type = self.current_instrument.projection_type

        # If we're in a dispersed mode, we need to know which axis the signal is dispersed along
        self.dispersion_axis = self.current_instrument.dispersion_axis()

        # Get the detector parameters (read noise, etc.)
        self.the_detector = self.current_instrument.the_detector

        # Initialize detector mask
        self.det_mask = 1.0

        # Get the background
        self.background = bg.Background(observation, webapp=webapp)

        # check if the saturation effect flag is True, otherwise turn saturation off by setting the fullwell
        # depth to np.inf (JETC-2944)
        if self.calculation_config.effects['saturation'] is not None:
            effect_saturation = self.calculation_config.effects['saturation']
        else:
            effect_saturation = self.the_detector.saturation
        self.the_detector.set_saturation(effect_saturation)

        # check if the readnoise flag is True, otherwise turn readnoise off by setting the value to 0
        if self.calculation_config.noise['readnoise'] is not None:
            noise_readnoise = self.calculation_config.noise['readnoise']
        else:
            noise_readnoise = self.the_detector.readnoise
        self.the_detector.set_readnoise(noise_readnoise)

        # check if the excess noise flag is True, otherwise turn excess noise off by setting the values to 0
        if self.calculation_config.noise['excess'] is not None:
            noise_excess = self.calculation_config.noise['excess']
        else:
            noise_excess = self.the_detector.excess
        self.the_detector.set_excess(noise_excess)

        # Then initialize the flux and wavelength grid
        astro.ConvolvedSceneCube.__init__(
            self,
            observation.scene,
            self.current_instrument,
            background=self.background,
            webapp=webapp,
            validate=validate,
            empty_scene=empty_scene
        )

        self.warnings.update(self.background.warnings)
        # We have to propagate the background through the system transmission to get the background in
        # e-/s/pixel/micron. The background rate is a 1D spectrum.
        self.bg_fp_rate = self.get_bg_fp_rate()
        # instantiate now so we can fill later
        self.total_rate = []

        self.projected_list = []
        # Loop over all slices and calculate the photon and electron rates through the observatory for each
        # one. Note that many modes (imaging, etc.) will have just a single slice.
        for flux_cube, flux_plus_bg in zip(self.flux_cube_list, self.flux_plus_bg_list):

            # Rates for the slice without the background
            projected_rate = self.all_rates(flux_cube, add_extended_background=False)

            # Rates for the slice with the background added
            projected_rate_plus_bg = self.all_rates(flux_plus_bg, add_extended_background=True)

            self.projected_list.append([projected_rate, projected_rate_plus_bg])

        # now compute the products of the 2D projection 
        self.signal_products()


    def slice_products(self, projected_rate, projected_rate_plus_bg):
        """
        Compute per-slice products from the projected 2D slice with post-projection detector effects added
        
        Parameters
        ----------
        projected_rate: dict
            A dictionary of 2D-projected signal from the rate cube.
        projected_rate_plus_bg: dict
            A dictionary of 2D-projected signal+background from the rate_plus_bg cube that will contain all
            other effects.
        """
        # apply detector effects to the projected products for this slice
        slice_rate = self.apply_detector_effects(projected_rate, add_flux_sources=False)
        slice_rate_plus_bg = self.apply_detector_effects(projected_rate_plus_bg, add_flux_sources=True)
        # Saturation map for the slice
        slice_saturation = self.the_detector._get_saturation_mask(slice_rate_plus_bg['fp_pix_no_ipc_unbinned'])
        slice_group = self.the_detector.get_before_sat(slice_rate_plus_bg['fp_pix_unbinned'])

        slice_saturation = self.binning.max(slice_saturation)
        slice_group = self.binning.min(slice_group)

        # The grid in the slice
        slice_pixgrid = self.get_pix_grid(slice_rate)

        # Append all slices to the master lists
        self.rate_list.append(slice_rate)
        self.rate_plus_bg_list.append(slice_rate_plus_bg)
        self.saturation_list.append(slice_saturation)
        self.groups_list.append(slice_group)
        self.pixgrid_list.append(slice_pixgrid)

    def signal_products(self):
        """
        Compute per-signal products from the projected 2D slice with post-projection detector effects added
        These will constitute the main interface and products used in DetectorSignal.
        
        """
        # Initialize slice lists
        self.rate_list = []
        self.rate_plus_bg_list = []
        self.saturation_list = []
        self.groups_list = []
        self.pixgrid_list = []

        for projected_rate, projected_rate_plus_bg in self.projected_list:
            self.slice_products(projected_rate, projected_rate_plus_bg)

        # the binning operator contains the new correct spatial grid
        self.grid_unbinned = self.grid
        self.grid = self.binning.grid
        try:
            self.det_mask = self.binning.dispersion(self.det_mask)
        except TypeError:
            pass

        # Get the mapping of wavelength to pixels on the detector plane. This is grabbed from the first entry
        # in self.rate_list and is currently defined to be the same for all slices.
        self.wave_pix = self.get_wave_pix()

        # This is also grabbed from the first slice as a diagnostic
        if self.validate:
            self.fp_rate = self.get_fp_rate()

        # Reassemble rates of multiple slices on the detector
        self.rate = self.on_detector(self.rate_list)
        self.rate_unbinned = self.on_detector(self.rate_list, product_name="fp_pix_unbinned")
        self.rate_no_qyield = self.on_detector(self.rate_list, product_name="fp_pix_no_qyield")
        self.rate_plus_bg = self.on_detector(self.rate_plus_bg_list)
        self.rate_plus_bg_unbinned = self.on_detector(self.rate_plus_bg_list, product_name="fp_pix_unbinned")

        # if len(rate_list) > 1, we have a detector image multiple slices tall.
        self.binning.cropshape[0] *= len(self.rate_list)
        self.binning.newshape[0] *= len(self.rate_list)

        # This code must be kept up-to-date with similar code in CombinedSignal.__init__
        # The bg_pix_rate 2D array will also have contributions from all the other detector effects that add
        # flux to the 2D projection.
        self.bg_pix_rate = self.get_bg_pix_rate()

        # Check to see if the background is saturating (on unbinned data)
        bgsat = self.get_saturation_mask(self.bg_pix_rate)
        if (np.sum(bgsat) > 0) or (np.isnan(np.sum(bgsat))):
            key = "background_saturated"
            self.warnings[key] = warning_messages[key]
        # bin to match the rest of the data
        self.bg_pix_rate = self.binning.sum(self.bg_pix_rate)

        self.ngroup_map = self.the_detector.get_before_sat(self.rate_plus_bg_unbinned)
        self.ngroup_map = self.binning.min(self.ngroup_map)
        self.fraction_saturation = np.max(self.the_detector.get_saturation_fraction(self.rate_plus_bg_unbinned))
        self.detector_pixels = self.current_instrument.get_detector_pixels(self.wave_pix)

        # JWST style
        self.brightest_pixel = np.max(self.rate_plus_bg)

        if isinstance(self.current_instrument, HSTInstrument):
            # HST Health and Safety.  Only applicable to HST instruments.
            self.brightest_pixel_rate = self.calc_brightest_pixel(self.rate_no_qyield,self.bg_pix_rate)
            self.detector_total_rate = self.calc_total_detector(self.rate_no_qyield,self.bg_pix_rate)

        # Get the read noise correlation matrix and store it as an attribute.
        if self.the_detector.rn_correlation:
            self.read_noise_correlation_matrix = self.current_instrument.get_readnoise_correlation_matrix(
                self.rate.shape)

    def spectral_detector_transform(self):
        """
        Create engine API format dict section containing properties of wavelength coordinates
        at the detector plane.

        Returns
        -------
        t: dict (engine API compliant keys)
        """
        t = {}
        t['wave_det_refpix'] = 0
        t['wave_det_max'] = self.wave_pix.max()
        t['wave_det_min'] = self.wave_pix.min()

        # there are currently five projection_type's which are basically detector plane types:
        #
        # 'spec' - where the detector plane is purely dispersion vs. spatial 
        # 'slitless' - basically a special case of 'spec' with where dispersion and spatial are mixed 
        # 'multiorder' - basically a special case of 'slitless' where there is more than one trace 
        # 'slitless_scan' - a special case of 'slitless' where the spectrum has been scanned in the 
        #     cross-dispersion direction by moving the telescope while exposing. 
        # 'image' - where the detector plane is purely spatial vs. spatial (i.e. no disperser element) 
        # 'image_scan' - a special case of 'image' where the image has been scanned in the Y direction 
        #     by moving the telescope while exposing.
        #
        # 'IFU' mode is of projection_type='spec' because the mapping from detector X pixels to wavelength is
        # the same for each slice.  this projection_type will work for 'MSA' mode as well because we will only
        # handle one aperture at a time.  'slitless' spectroscopy will mix spatial and dispersion information
        # onto the detector X axis. However, the detector plane is fundamentally spatial vs. wavelength in
        # that case so it's handled the same as projection_type='spec'. creating a spectrum for a specific
        # target will be handled via the extraction strategy.
        if self.projection_type in ('spec', 'slitless', 'slitless_scan', 'multiorder'):
            t['wave_det_size'] = len(self.wave_pix)
            if len(self.wave_pix) > 1:
                # we don't yet have a way of handling non-linear coordinate transforms here. that said, this
                # is mostly right for most of our cases with nirspec prism being the notable exception. this
                # is also only used for plotting purposes while the true actual wave_pix mapping is used
                # internally for all calculations.
                t['wave_det_step'] = (self.wave_pix[-1] - self.wave_pix[0]) / t['wave_det_size']
            else:
                t['wave_det_step'] = 0.0
            t['wave_det_refval'] = self.wave_pix[0]
        elif self.projection_type in ("image", "image_scan"):
            t['wave_det_step'] = 0.0
            t['wave_det_refval'] = self.wave_pix[0]
            t['wave_det_size'] = 1
        else:
            message = "Unsupported projection_type: %s" % self.projection_type
            raise EngineOutputError(value=message)
        return t

    def wcs_info(self):
        """
        Get detector coordinate transform as a dict of WCS keyword/value pairs.

        Returns
        -------
        header: dict
            WCS header keys defining coordinate transform in the detector plane
        """
        if self.projection_type in ('image', 'image_scan'):
            # if we're in imaging mode, the detector sampling is the same as the model
            header = self.grid.wcs_info()
        elif self.projection_type in ('spec', 'slitless', 'slitless_scan', 'multiorder'):
            # if we're in a dispersed mode, dispersion can be either along the X or Y axis. the image outputs in
            # the engine Report are rotated so that dispersion will always appear to be along the X axis with
            # wavelength increasing with increasing X (i. e. dispersion angle of 0).  currently, the only other
            # supported dispersion angle is 90 which is what we get when dispersion_axis == 'y'.
            t = self.grid.as_dict()
            t.update(self.spectral_detector_transform())
            header = {
                'ctype1': 'Wavelength',
                'crpix1': 1,
                'crval1': t['wave_det_min'] - 0.5 * t['wave_det_step'],
                'cdelt1': t['wave_det_step'],
                'cunit1': 'um',
                'cname1': 'Wavelength',
                'ctype2': 'Y offset',
                'crpix2': 1,
                'crval2': t['y_min'] - 0.5 * t['y_step'],
                'cdelt2': -t['y_step'],
                'cunit2': 'arcsec',
                'cname2': 'Detector Offset',
            }
            if self.dispersion_axis == 'y':
                header['ctype2'] = 'X offset'
                header['crval2'] = t['x_min'] - 0.5 * t['x_step'],
                header['cdelt2'] = t['x_step']
        else:
            message = "Unsupported projection_type: %s" % self.projection_type
            raise EngineOutputError(value=message)
        return header

    def get_wave_pix(self):
        """
        Return the mapping of wavelengths to pixels on the detector plane
        """
        return self.rate_list[0]['wave_pix']

    def get_fp_rate(self):
        """
        Return scene flux at the focal plane in e-/s/pixel/micron (excludes background)
        """
        if not self.validate:
            raise EngineOutputError("Focal Plane Rate Cube unavailable. Recompute with validate=True.")
        return self.rate_list[0]['fp']

    def get_bg_fp_rate(self):
        """
        Calculate background in e-/s/pixel/micron at the focal plane. Also correct for any excess in predicted
        background if there are pupil losses in the PSF. (#2529)

        Returns
        -------
        bg_fp_rate: np.ndarray
            The background 1D array converted to e-/s/pixel/micron
        """
        bg_fp_rate = self.focal_plane_rate(self.ote_rate(self.background.mjy_pix), self.wave)
        wave_range = self.current_instrument.get_wave_range()
        pupil_thru = self.current_instrument.psf_library.get_pupil_throughput(wave_range['wmin'],
                                                                              self.current_instrument.instrument[
                                                                                  'instrument'],
                                                                              self.current_instrument.instrument[
                                                                                  'aperture'])
        return bg_fp_rate * pupil_thru

    def get_bg_pix_rate(self):
        """
        Calculate the background on the detector in e-/s/pixel

        By computing off of these on-detector stitched properties, we can check if ANY of the background is
        saturating.

        Returns
        -------
        bg_pix_rate: numpy.ndarray
            The 2D unbinned background-only pixel rate
        """
        bg_pix_rate = self.rate_plus_bg_unbinned - self.rate_unbinned
        return bg_pix_rate

    def on_detector(self, rate_list, product_name="fp_pix"):
        """
        This will take the list of (pixel) rates and use them create a single detector frame. A single image
        will only have one rate in the list, but the IFUs will have n_slices. There may be other examples,
        such as different spectral orders for NIRISS. It is not yet clear how many different flavors there
        are, so this step may get refactored if it gets too complicated. Observing modes that only have one
        set of rates (imaging and single-slit spectroscopy, for instance) will still go through this, but the
        operation is trivial.

        Returns
        -------
        A 2D plane with any stripes added together into a final product.
        """
        aperture_sh = rate_list[0][product_name].shape
        n_apertures = len(rate_list)
        detector_shape = (aperture_sh[0] * n_apertures, aperture_sh[1])
        detector = np.zeros(detector_shape, dtype=np.float32)

        i = 0
        for rate in rate_list:
            detector[i * aperture_sh[0]:(i + 1) * aperture_sh[0], :] = rate[product_name]
            i += 1

        return detector

    def get_pix_grid(self, rate):
        """
        Generate the coordinate grid of the detector plane
        """
        if self.projection_type in ('image', 'image_scan'):
            grid = self.grid
        elif self.projection_type in ('spec', 'slitless', 'slitless_scan', 'multiorder'):
            nw = rate['wave_pix'].shape[0]
            if self.dispersion_axis == 'x':
                # for slitless calculations, the dispersion axis is longer than the spectrum being dispersed
                # because the whole field of view is being dispersed. 'excess' is the size of the FOV and half
                # will be to the left of the blue end of the spectrum and half to the right of the red end.
                # this is used to create the new spatial coordinate transform for the pixel image on the
                # detector.
                excess = rate['fp_pix'].shape[1] - nw
                pix_grid = coords.IrregularGrid(
                    self.grid.col,
                    (np.arange(nw + excess) - (nw + excess) / 2.0) * self.grid.xsamp
                )
            else:
                excess = rate['fp_pix'].shape[0] - nw
                pix_grid = coords.IrregularGrid(
                    (np.arange(nw + excess)[::-1] - (nw + excess) / 2.0) * self.grid.ysamp,
                    self.grid.row
                )
            return pix_grid
        else:
            raise EngineOutputError(value="Unsupported projection_type: %s" % self.projection_type)
        return grid

    def export_fp_to_fits(self, fitsfile='fpcube'):
        """
        Write focal plane rate cube to a FITS file
        """
        header = self.grid.wcs_info()
        slice_index=0

        if self.validate:
            fitsfile_slice = '{}{}.fits'.format(fitsfile, str(slice_index).strip())
            t1 = fits.PrimaryHDU(np.moveaxis(self.fp_rate, [0, 1, 2], [1, 2, 0]))
            c1 = fits.Column(name='wavelength', array=self.wave, format='D')
            t2 = fits.BinTableHDU.from_columns([c1])
            t1.header.update(header)
            tbhdu = fits.HDUList([t1,t2])
            tbhdu.writeto(fitsfile_slice, overwrite=True)
            tbhdu.close()
        else:
            raise EngineOutputError("Focal Plane Rate Cube unavailable. Recompute with validate=True.")

    def _build_dark_array(self, wave, size_y):
        '''
        Builds the dark current array, in case there is more than one value of dark current. Case in point:
        segmented detectors such as COS FUV.

        This operates by replacing the dark_current value in the detector instance, originally a dict, with a
        2D array with the proper dark values filling their corresponding slices on the array.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength array.
        size_y: int
            Size of array in the Y direction
        '''

        # Code below assumes that we are in a COS FUV spectral calculation.
        # It may turn out to be more generic than that, but we will only know
        # when actual test cases become available.

        # check if the dark current effect flag is True, otherwise turn dark current off by setting the value to 0
        if self.calculation_config.noise['dark'] is not None:
            noise_dark = self.calculation_config.noise['dark']
        else:
            noise_dark = self.the_detector.dark
        if not noise_dark:
            self.the_detector.dark_current = 0.0


        if (hasattr(self.the_detector, 'segnames') or hasattr(self.the_detector, 'stripenames')) and \
            isinstance(self.the_detector.dark_current, dict):

            # build empty dark array
            nx = wave.shape[0]
            ny = size_y
            dark_current = np.zeros(shape=(ny, nx))

            segment_indices = self.current_instrument.get_segment_indicies(wave)
            # loop over each segment
            for segment in segment_indices:

                # build array with dark values for segment
                nx_slice = segment['wmax_ind'] - segment['wmin_ind']
                dark_slice = np.full((ny, nx_slice), self.the_detector.dark_current[segment['segname']])

                # add segment to dark array
                dark_current[::, segment['wmin_ind']:segment['wmax_ind']] = dark_slice

            # replace dark in detector instance with the dark array.
            self.the_detector.dark_current = dark_current

    def all_rates(self, flux, add_extended_background=False):
        """
        Calculate rates in e-/s/pixel/micron or e-/s/pixel given a flux cube in mJy This step carries out the
        projection of the flux cube into a 2D detector signal

        Parameters
        ----------
        flux: ConvolvedSceneCube instance
            Convolved source flux cube with flux units in mJy
        add_extended_background: bool (default=False)
            Toggle for including extended background not contained within the flux cube but present on the 2D
            projection.

        Returns
        -------
        projected: dict
            The result of the 2D projections, bundled into a dictionary
        """
        # The source rate at the telescope aperture
        ote_rate = self.ote_rate(flux)

        # The source rate at the focal plane in interacting photons/s/pixel/micron
        fp_rate = self.focal_plane_rate(ote_rate, self.wave)

        self.total_rate.append(np.asarray(fp_rate.sum(axis=(0,1))))

        # the fp_pix_variance is the variance of the per-pixel electron rate and includes the chromatic effects
        # of quantum yield.
        if self.projection_type == 'image':
            # The wavelength-integrated rate in e-/s/pixel, relevant for imagers
            fp_pix_rate, fp_pix_no_qyield, fp_pix_variance = self.image_rate(fp_rate)
            wave_pix = self.wave_eff(fp_rate)

        elif self.projection_type == 'spec':
            # The wavelength-integrated rate in e-/s/pixel, relevant for spectroscopy
            wave_pix, fp_pix_rate, fp_pix_no_qyield, fp_pix_variance = self.spec_rate(fp_rate)

        elif self.projection_type in ('slitless', 'multiorder'):
            # The wavelength-integrated rate in e-/s/pixel, relevant for slitless spectroscopy
            wave_pix, fp_pix_rate, fp_pix_no_qyield, fp_pix_variance = self.slitless_rate(
                fp_rate,
                add_extended_background=add_extended_background
            )
        elif self.projection_type == 'slitless_scan':
            # The wavelength-integrated rate in e-/s/pixel, relevant for slitless spectroscopy
            wave_pix, fp_pix_rate, fp_pix_no_qyield, fp_pix_variance = self.slitless_scan_rate(
                fp_rate,
                add_extended_background=add_extended_background
            )
        elif self.projection_type == 'image_scan':
            # The wavelength-integrated rate in e-/s/pixel, relevant for slitless spectroscopy
            wave_pix, fp_pix_rate, fp_pix_no_qyield, fp_pix_variance = self.image_scan_rate(
                fp_rate,
                add_extended_background=add_extended_background
            )

        else:
            raise EngineOutputError(value="Unsupported projection_type: %s" % self.projection_type)

        products = {"wave_pix": wave_pix, "fp_pix_rate": fp_pix_rate, "fp_pix_no_qyield": fp_pix_no_qyield, "fp_pix_variance": fp_pix_variance}
        if self.validate:
            products["ote_rate"] = ote_rate
            products["fp_rate"] =  fp_rate

        return products

    def apply_detector_effects(self, projected, add_flux_sources=False):
        """
        This part of the code accepts projected data and applies subsequent 2D detector effects. Some of them
        add additional flux to the array and should only be applied to _plus_bg products. 
        - Dark current 
        - CCD Post-flash electrons 
        - WFC3 Thermal Background 
        - IPC 
        - Detector Binning
        

        Parameters
        ----------
        projected: dict
            The result of the 2D projections, bundled into a dictionary
        add_flux_sources: bool
            Boolean controlling whether we add additional sources of flux from detector effects to the data. 

        Returns
        -------
        products: dict
            Dict of products produced by rate calculation.
                'wave_pix' - Mapping of wavelength to detector pixels 
                'ote' - Source rate at the telescope aperture 
                'fp' - Source rate at the focal plane in e-/s/pixel/micron 
                'fp_pix' - Source rate per pixel 
                'fp_pix_no_ipc' - Source rate per pixel excluding effects of 
                      inter-pixel capacitance
                'fp_pix_no_qyield' - Source rate per pixel, not adjusted for 
                      quantum yield (still in incident photons/s/pixel/micron)
            
        """
        #unpack
        wave_pix = projected["wave_pix"]
        fp_pix_rate = copy.deepcopy(projected["fp_pix_rate"])
        fp_pix_variance = copy.deepcopy(projected["fp_pix_variance"])
        fp_pix_no_qyield = projected["fp_pix_no_qyield"]
        # These are not altered, therefore no copy is needed
        if self.validate:
            ote_rate = projected["ote_rate"] 
            fp_rate = projected["fp_rate"]

        # if dark current depends on position on the detector, build it as a 2D array
        # (case in point: COS FUV spec)
        self._build_dark_array(wave_pix, fp_pix_rate.shape[0])

        # Dark should be added if and only if we're processing a rate with a background
        if add_flux_sources:
            self.additional_flux = self.the_detector.dark_current
            if self.current_instrument.inst_name == "wfc3":
                thermal = self.current_instrument.add_thermal()
                self.additional_flux += thermal
            if "postflash" in self.the_detector.input_detector:
                # postflash is a discrete value per exposure, but we need it to be a rate here.
                self.additional_flux += self.the_detector.postflash_as_rate()
            fp_pix_rate += self.additional_flux
            fp_pix_variance += self.additional_flux

        # Include IPC effects, if available and requested
        if self.the_detector.ipc:
            kernel = self.current_instrument.get_ipc_kernel()
            fp_pix_rate_ipc = self.ipc_convolve(fp_pix_rate, kernel)
        else:
            fp_pix_rate_ipc = copy.deepcopy(fp_pix_rate)

        fp_pix_unbinned = fp_pix_rate_ipc
        fp_pix_no_ipc_unbinned = fp_pix_rate
        fp_pix_no_qyield_unbinned = fp_pix_no_qyield
        fp_pix_variance_unbinned = fp_pix_variance
        wave_pix = self.binning.dispersion(wave_pix)
        fp_pix_rate_ipc = self.binning.sum(fp_pix_rate_ipc)
        fp_pix_rate = self.binning.sum(fp_pix_rate)
        fp_pix_no_qyield = self.binning.sum(fp_pix_no_qyield)
        fp_pix_variance = self.binning.sum(fp_pix_variance)

        # fp_pix is the final product. Since there is no reason to
        # carry around the ipc label everywhere, we rename it here.
        products = {
            'wave_pix': wave_pix,
            'fp_pix': fp_pix_rate_ipc,
            'fp_pix_no_ipc': fp_pix_rate,  # this is for calculating saturation
            'fp_pix_no_qyield': fp_pix_no_qyield, # this is for calculating feasibility thread
            'fp_pix_variance': fp_pix_variance,  # this is for calculating the detector noise
            'fp_pix_unbinned': fp_pix_unbinned, # this is fp_pix before binning
            'fp_pix_no_ipc_unbinned': fp_pix_no_ipc_unbinned, # this is fp_pix_no_ipc before binning
            'fp_pix_no_qyield_unbinned': fp_pix_no_qyield_unbinned, # this is calculating feasibility thread 
            'fp_pix_variance_unbinned': fp_pix_variance_unbinned  # this is for calculating the detector noise
        }
        if self.validate:
            products['ote'] = ote_rate
            products['fp'] = fp_rate
        return products

    def spec_rate(self, rate):
        '''
        For slitted spectrographs, calculate the detector signal by integrating along the dispersion direction
        of the cube (which is masked by a, by assumption, narrow slit). For slitless systems or slits wider
        than the PSF, the slitless_rate method should be used to preserve spatial information within the slit.

        Parameters
        ---------
        rate: numpy.ndarray
            Rate of photons interacting with detector as a function of model wavelength set

        Returns
        -------
        products: 4-element tuple of numpy.ndarrays
            first element - map of pixel to wavelength 
            second element - electron rate per pixel 
            third element - photon rate per pixel (no qyield
            fourth element - variance of electron rate per pixel
        '''
        dispersion = self.current_instrument.get_dispersion(self.wave)
        wave_pix = self.current_instrument.get_wave_pix()
        wave_pix_trunc = wave_pix[np.where(np.logical_and(wave_pix >= self.wave.min(),
                                                          wave_pix <= self.wave.max()))]

        # Check that the source spectrum is actually inside the instrumental wavelength
        # coverage.
        if len(wave_pix_trunc) == 0:
            raise RangeError(value='wave and wave_pix do not overlap')

        # Check the dispersion axis to determine which axis to sum and interpolate over
        if self.dispersion_axis == 'x':
            axis = 1
        else:
            axis = 0

        # We can simply sum over the dispersion direction. This is where we lose the spatial information within the aperture.
        spec_rate = np.sum(rate, axis=axis)

        # And then scale to the dispersion function (pixel/micron) to transform
        # from e-/s/micron to e-/s/pixel.
        spec_rate_pix = spec_rate * dispersion

        # but we are still sampled on the internal grid, so we have to interpolate to the pixel grid.
        # use kind='slinear' since it's ~2x more memory efficient than 'linear'. 'slinear' uses different code path to
        # calculate the slopes.
        # at this point, whether you used the x (1) or y (0) axis to sum over, the wavelength axis is now 1.

        int_spec_rate = sci_int.interp1d(self.wave, spec_rate_pix, axis=1, kind='slinear', assume_sorted=True,
                                         copy=False)
        spec_rate_pix_sampled = int_spec_rate(wave_pix_trunc)

        # Handle a detector gap here by constructing a mask. If the current_instrument implements it,
        # it'll be a real mask array.  Otherwise it will simply be 1.0.
        self.det_mask = self.current_instrument.create_gap_mask(wave_pix_trunc)

        # this is the interacting photon rate in the detector with mask applied.
        spec_rate_pix_sampled *= self.det_mask

        # Add effects of non-unity quantum yields. For the spec projection, we assume that the quantum yield does not
        # change over a spectral element. Then we can just multiply the products by the relevant factors.
        q_yield, fano_factor = self.current_instrument.get_quantum_yield(wave_pix_trunc)

        # convert the photon rate to electron rate by multiplying by the quantum yield which is a function of wavelength
        spec_electron_rate_pix = spec_rate_pix_sampled * q_yield
        spec_rate_no_qyield = spec_rate_pix_sampled

        # to meet IDT expectations, some instruments require a possibly chromatic fudge factor to be applied
        # to the per-pixel electron rate variance.
        var_fudge = self.current_instrument.get_variance_fudge(wave_pix_trunc)

        # the variance in the electron rate, Ve, is also scaled by the quantum yield plus a fano factor which is
        # analytic in the simple 1 or 2 electron case: Ve = (qy + fano) * Re.  since Re is the photon rate
        # scaled by the quantum yield, Re = qy * Rp, we get: Ve = qy * (qy + fano) * Rp
        spec_electron_variance_pix = spec_rate_pix_sampled * q_yield * (q_yield + fano_factor) * var_fudge

        # bin the instrument wavelength grid in pixels
        if self.dispersion_axis == 'x':
            bin_x = self.current_instrument.get_dispersion_binning()
            bin_y = self.current_instrument.get_spatial_binning()
        else:
            spec_electron_rate_pix = np.rot90(spec_electron_rate_pix)
            spec_electron_variance_pix = np.rot90(spec_electron_variance_pix)
            spec_rate_no_qyield = np.rot90(spec_rate_no_qyield)
            wave_pix_trunc = wave_pix_trunc[::-1]
            bin_x = self.current_instrument.get_spatial_binning()
            bin_y = self.current_instrument.get_dispersion_binning()
        self.binning = DetectorBinning(spec_electron_rate_pix, bin_x, bin_y, self.grid, dispersion_axis=self.dispersion_axis, wave=wave_pix_trunc)

        products = wave_pix_trunc, spec_electron_rate_pix, spec_rate_no_qyield, spec_electron_variance_pix

        return products

    def image_rate(self, rate):
        '''
        Calculate the electron rate for imaging modes by integrating along the wavelength direction of the
        cube.

        Parameters
        ---------
        rate: numpy.ndarray
            Rate of photons interacting with detector as a function of model wavelength set

        Returns
        -------
        products: 3-element tuple of numpy.ndarrays
            first element - electron rate per pixel
            second element - photon rate per pixel (no quantum yield)
            third element - variance of electron rate per pixel
        '''
        q_yield, fano_factor = self.current_instrument.get_quantum_yield(self.wave)

        # convert the photon rate to electron rate by multiplying by the quantum yield which is a function of wavelength
        electron_rate_pix = integrate.trapz(rate * q_yield, self.wave)
        image_rate_no_qyield = integrate.trapz(rate, self.wave)

        # to meet IDT expectations, some instruments require a possibly chromatic fudge factor to be applied
        # to the per-pixel electron rate variance.
        var_fudge = self.current_instrument.get_variance_fudge(self.wave)

        # the variance in the electron rate, Ve, is also scaled by the quantum yield plus a fano factor which is
        # analytic in the simple 1 or 2 electron case: Ve = (qy + fano) * Re.  since Re is the photon rate
        # scaled by the quantum yield, Re = qy * Rp, we get: Ve = qy * (qy + fano) * Rp
        electron_variance_pix = integrate.trapz(rate * q_yield * (q_yield + fano_factor) * var_fudge, self.wave)

        # bin the instrument wavelength grid in pixels
        bin_x = self.current_instrument.get_spatial_binning()
        bin_y = self.current_instrument.get_spatial_binning()
        self.binning = DetectorBinning(electron_rate_pix, bin_x, bin_y, self.grid)

        products = electron_rate_pix, image_rate_no_qyield, electron_variance_pix

        return products

    def slitless_rate(self, rate, add_extended_background=True):
        '''
        Calculate the detector rates for slitless modes. Here we retain all spatial information and build up
        the detector plane by shifting and coadding the frames from the convolved flux cube. Also need to
        handle and add background that comes from outside the flux cube, but needs to be accounted for.

        How this works: 
        1. We calculate wave_pix and truncate it to the wavelength range of the instrument (which should 
        already have been done, but it is a check) 
        2. We grab the detector_pixel (dispersion) and trace (cross-dispersion) pixel locations, 
        interpolated per wavelength on the detector. Except for modes with specific tracefile that map 
        wavelength to (detector_pixel,trace), these will simply be numerical and constant-0 files specifying
        the upper left corner of where to put the wavelength grid.
        3. We then step through wavelengths. If there IS a trace file, we shift in the cross dispersion
        direction and add the rate cube plane to the image at that location. Otherwise, no expensive 
        shifting operation is necessary. 
        4. If extended background is switched on, we add background for that wavelength to the rest of the 
        detector image not covered by the rate cube plane, because background from all wavelengths will fall 
        on all parts of a slitless calculation.

        Parameters
        ----------
        rate: 3D numpy.ndarray
            Cube containing the flux rate at the focal plane
        add_extended_background: bool (default: True)
            Toggle for including extended background not contained within the flux cube

        Returns
        -------
        products: 4 entry tuple
            wave_pix: 1D numpy.ndarray containing wavelength to pixel mapping on the detector plane 
            spec_rate: 2D numpy.ndarray of detector count rates 
            spec_no_qyield: 2D numpy.ndarray of the detector rate without quantum yield correction 
            spec_variance: 2D numpy.ndarray of the variance array.
        '''
        wave_pix = self.current_instrument.get_wave_pix()
        wave_subs = np.where(
            np.logical_and(
                wave_pix >= self.wave.min(),
                wave_pix <= self.wave.max()
            )
        )
        wave_pix_trunc = wave_pix[wave_subs]

        if len(wave_pix_trunc) == 0:
            raise RangeError(value='wave and wave_pix do not overlap')

        dispersion = self.current_instrument.get_dispersion(wave_pix_trunc)
        trace = self.current_instrument.get_trace(wave_pix_trunc)
        dpix = self.current_instrument.get_detector_pixels(wave_pix_trunc).astype(int)

        if dispersion.shape != wave_pix_trunc.shape:
            raise DataConfigurationError("Dispersion and wavelength axes do not match! This should never happen.")

        q_yield, fano_factor = self.current_instrument.get_quantum_yield(wave_pix_trunc)
        # if we kind='slinear' since it's ~2x more memory efficient than 'linear'. 'slinear' uses different code
        # path to calculate the slopes. However, slinear is *much* slower, so it is a tradeoff. Also lowering the
        # rate type to float32 to conserve memory.
        int_rate_pix = sci_int.interp1d(self.wave, rate.astype(np.float32, casting='same_kind'),
                                        kind='linear', axis=2, assume_sorted=True, copy=False)
        rate_pix = int_rate_pix(wave_pix_trunc)

        # convert the photon rate to electron rate by multiplying by the quantum yield which is a function of
        # wavelength
        electron_rate_pix = rate_pix * q_yield
        electron_rate_no_qyield = rate_pix

        # to meet IDT expectations, some instruments require a possibly chromatic fudge factor to be applied
        # to the per-pixel electron rate variance.
        var_fudge = self.current_instrument.get_variance_fudge(wave_pix_trunc)

        # the variance in the electron rate, Ve, is also scaled by the quantum yield plus a fano factor which
        # is analytic in the simple 1 or 2 electron case: Ve = (qy + fano) * Re.  since Re is the photon rate
        # scaled by the quantum yield, Re = qy * Rp, we get: Ve = qy * (qy + fano) * Rp
        electron_variance_pix = rate_pix * q_yield * (q_yield + fano_factor) * var_fudge

        # interpolate the background onto the pixel spacing
        int_bg_fp_rate = sci_int.interp1d(self.wave, self.bg_fp_rate.astype(np.float32, casting='same_kind'),
                                          kind='linear', assume_sorted=True, copy=False)
        bg_fp_rate_pix = int_bg_fp_rate(wave_pix_trunc)

        # calculate electron rate and variance due to background
        bg_electron_rate = bg_fp_rate_pix * q_yield
        bg_electron_variance = bg_fp_rate_pix * q_yield * (q_yield + fano_factor) * var_fudge

        # The first part of this code is meant to add the PSF images from the convolved scene cube along
        # either the x or y axis depending on the dispersion axis, optionally following the path of a spectral
        # trace (currently used only for SOSS mode). The psfs will be added to all locations within the
        # resolution element.
        #
        # Because, in slitless modes, the disperser is dispersing light coming in from everywhere in the pupil
        # plane, every part of the detector should have a contribution from every wavelength of light (from
        # both orders, for SOSS mode). The add_extended_background statement does that - it fills every pixel
        # up to i, and after i+rate_pix.shape[1], with the same background that comes baked into the rate_pix
        # images thanks to the AdvancedPSF functions that create the convolved scene cube.

        if self.empty_scene:
            # if we have an explicitly empty scene, we're doing a background-only order calculation and don't
            # need to even pretend to disperse the spectrum - CombinedSignal will handle padding it to match
            # the interesting order(s) and this way there will be no need to trim.
            spec_shape = (rate_pix.shape[0], rate_pix.shape[1])
            spec_rate = np.zeros(spec_shape)
            spec_rate_no_qyield = np.zeros(spec_shape)
            spec_variance = np.zeros(spec_shape)
            if add_extended_background:
                for i in np.arange(dispersion.shape[0]):
                    spec_rate += bg_electron_rate[i] * dispersion[i]
                    spec_rate_no_qyield += bg_fp_rate_pix[i] * dispersion[i]
                    spec_variance += bg_electron_variance[i] * dispersion[i]
        else:  # if the scene is data
            # dispersion_axis tells us whether we need to sum the planes of the cube horizontally
            # or vertically on the detector plane.
            if self.dispersion_axis == 'x':
                spec_shape = (rate_pix.shape[0], rate_pix.shape[2] + rate_pix.shape[1] + dpix[0])
                spec_rate = np.zeros(spec_shape)
                spec_rate_no_qyield = np.zeros(spec_shape)
                spec_variance = np.zeros(spec_shape)
                for i in range(len(wave_pix_trunc)):
                    # Background not yet completely added. Make sure there is a trace shift to be done so that
                    # we don't make an expensive call to shift() if we don't have to. Use mode='nearest' to
                    # fill in new pixels with background when image is shifted. 
                    # dispersion pixel location for any given wavelength is dpix, cross-dispersion pixel
                    # location is trace
                    if trace[i] != 0.0:
                        spec_rate[:, dpix[i]:dpix[i] + rate_pix.shape[1]] += shift(
                            electron_rate_pix[:, :, i],
                            shift=(trace[i], 0),
                            mode='nearest',
                            order=1
                        ) * dispersion[i]
                        spec_rate_no_qyield[:, dpix[i]:dpix[i] + rate_pix.shape[1]] += shift(
                            electron_rate_no_qyield[:, :, i],
                            shift=(trace[i], 0),
                            mode='nearest',
                            order=1
                        ) * dispersion[i]
                        spec_variance[:, dpix[i]:dpix[i] + rate_pix.shape[1]] += shift(
                            electron_variance_pix[:, :, i],
                            shift=(trace[i], 0),
                            mode='nearest',
                            order=1
                        ) * dispersion[i]
                    else:
                        spec_rate[:, dpix[i]:dpix[i] + rate_pix.shape[1]] += electron_rate_pix[:, :, i] * dispersion[i]
                        spec_rate_no_qyield[:, dpix[i]:dpix[i] + rate_pix.shape[1]] += electron_rate_no_qyield[:, :, i] * dispersion[i]
                        spec_variance[:, dpix[i]:dpix[i] + rate_pix.shape[1]] += electron_variance_pix[:, :, i] * dispersion[i]

                    # Adding background to all other pixels, unless we are asked not to.
                    if add_extended_background:
                        spec_rate[:, :dpix[i]] += bg_electron_rate[i] * dispersion[i]
                        spec_rate[:, dpix[i] + rate_pix.shape[1]:] += bg_electron_rate[i] * dispersion[i]
                        spec_rate_no_qyield[:, :dpix[i]] += bg_fp_rate_pix[i] * dispersion[i]
                        spec_rate_no_qyield[:, dpix[i] + rate_pix.shape[1]:] += bg_fp_rate_pix[i] * dispersion[i]
                        spec_variance[:, :dpix[i]] += bg_electron_variance[i] * dispersion[i]
                        spec_variance[:, dpix[i] + rate_pix.shape[1]:] += bg_electron_variance[i] * dispersion[i]
            else:  # if the dispersion is on the y axis
                spec_shape = (rate_pix.shape[2] + rate_pix.shape[0] + dpix[0], rate_pix.shape[1])
                spec_rate = np.zeros(spec_shape)
                spec_rate_no_qyield = np.zeros(spec_shape)
                spec_variance = np.zeros(spec_shape)
                for i in range(len(wave_pix_trunc)):
                    # Background not yet completely added. Make sure there is a trace shift to be done so that
                    # we don't make an expensive call to shift() if we don't have to. Use mode='nearest' to
                    # fill in new pixels with background when image is shifted.
                    # dispersion pixel location for any given wavelength is dpix, cross-dispersion pixel
                    # location is trace
                    if trace[i] != 0.0:
                        spec_rate[dpix[i]:dpix[i] + rate_pix.shape[0], :] += shift(
                            electron_rate_pix[:, :, i],
                            shift=(0, trace[i]),
                            mode='nearest',
                            order=1
                        ) * dispersion[i]
                        spec_rate_no_qyield[dpix[i]:dpix[i] + rate_pix.shape[0], :] += shift(
                            electron_rate_no_qyield[:, :, i],
                            shift=(0, trace[i]),
                            mode='nearest',
                            order=1
                        ) * dispersion[i]
                        spec_variance[dpix[i]:dpix[i] + rate_pix.shape[0], :] += shift(
                            electron_variance_pix[:, :, i],
                            shift=(0, trace[i]),
                            mode='nearest',
                            order=1
                        ) * dispersion[i]
                    else:
                        spec_rate[dpix[i]:dpix[i] + rate_pix.shape[0], :] += electron_rate_pix[:, :, i] * dispersion[i]
                        spec_rate_no_qyield[dpix[i]:dpix[i] + rate_pix.shape[0], :] += electron_rate_no_qyield[:, :, i] * dispersion[i]
                        spec_variance[dpix[i]:dpix[i] + rate_pix.shape[0], :] += electron_variance_pix[:, :, i] * dispersion[i]
                    # Adding background to all other pixels, unless we are asked not to.
                    if add_extended_background:
                        spec_rate[:dpix[i], :] += bg_electron_rate[i] * dispersion[i]
                        spec_rate[dpix[i] + rate_pix.shape[0]:, :] += bg_electron_rate[i] * dispersion[i]
                        spec_rate_no_qyield[:dpix[i], :] += bg_fp_rate_pix[i] * dispersion[i]
                        spec_rate_no_qyield[dpix[i] + rate_pix.shape[0]:, :] += bg_fp_rate_pix[i] * dispersion[i]
                        spec_variance[:dpix[i], :] += bg_electron_variance[i] * dispersion[i]
                        spec_variance[dpix[i] + rate_pix.shape[0]:, :] += bg_electron_variance[i] * dispersion[i]

        # bin the instrument wavelength grid in pixels
        if self.dispersion_axis == 'x':
            bin_x = self.current_instrument.get_dispersion_binning()
            bin_y = self.current_instrument.get_spatial_binning()
        else:
            bin_x = self.current_instrument.get_spatial_binning()
            bin_y = self.current_instrument.get_dispersion_binning()
        self.binning = DetectorBinning(spec_rate, bin_x, bin_y, self.grid, dispersion_axis=self.dispersion_axis, wave=wave_pix_trunc, projection_type=self.projection_type)

        # dispersion_axis determines whether wavelength is the first or second axis
        if self.dispersion_axis == 'x' or self.projection_type == 'multiorder':
            products = wave_pix_trunc, spec_rate, spec_rate_no_qyield, spec_variance
        else:
            # if dispersion is along Y, wavelength increases bottom to top, but Y index increases top to bottom.
            # flip the Y axis to account for this.
            products = wave_pix_trunc, np.flipud(spec_rate), np.flipud(spec_rate_no_qyield), np.flipud(spec_variance)

        return products


    def observation_scan(self, product, bgvals, add_extended_background=True):
        '''
        Calculate the scan effect. We build up the actual detector plane by shifting and adding an appropriate 
        amount of the previously-projected image, and then rebuild the Grid.

        How this works:
        1. We calculate the scan properties
        2. We calculate the actual scan length in pixels.
        3. We divide the existing products by that length, as it will be divided out into those pixels.
        4. We shift and add the product to the new array
        5. If add_extended_background = True, we add the background to the parts of the new array not covered 
           by the shifted product.

        Parameters
        ----------
        products: numpy.ndarray
            Output of one of the base projections
        add_extended_background: bool (default: True)
            Toggle for including extended background not contained within the flux cube

        Returns
        -------
        products: 4 entry tuple, a scanned version of the input tuple.
        '''
        # Compute scan properties. The get_scan_length() function is only run here.
        scan_length = self.current_instrument.get_scan_length()
        scan_angle = self.the_detector.input_detector.get("scan_angle",0)*np.pi/180.

        if self.dispersion_axis == 'x':
            # the scan length will rarely be integer pixels
            scan_length_pix = scan_length / self.current_instrument.get_aperture_pars()['plate_scale'][0] #pixscl_y
            dpix = np.arange(int(np.ceil(scan_length_pix)))
            # scan must hold the additional half-full pixel
            scan_shape = (product.shape[0] + int(np.ceil(scan_length_pix)), product.shape[1]) # get from the 2D signal
            scan_product = np.zeros(scan_shape)
            axis = 0
        else:
            # the scan length will rarely be integer pixels
            scan_length_pix = scan_length / self.current_instrument.get_aperture_pars()['plate_scale'][1] #pixscl_x
            dpix = np.arange(np.ceil(scan_length_pix))
            # scan must hold the additional half-full pixel
            scan_shape = (product.shape[0], product.shape[1] + int(np.ceil(scan_length_pix))) # get from the 2D signal
            scan_product = np.zeros(scan_shape)
            axis = 1

        # set up the relevant slices
        iter_piece = [slice(dpix[i],dpix[i]+product.shape[axis]) for i in dpix]
        iter_all = [slice(None) for i in dpix]
        iter_before = [slice(0,dpix[i]) for i in dpix]
        iter_after = [slice(dpix[i]+product.shape[axis], None) for i in dpix]

        # which axes should the slices be applied to?
        # has to be set after dpix is defined AND after the list comprehensions
        if self.dispersion_axis == "x":
            slicelist = [iter_piece, iter_all]
            slicebefore = [iter_before, iter_all]
            sliceafter = [iter_after, iter_all]
        else:
            slicelist = [iter_all, iter_piece]
            slicebefore = [iter_all, iter_before]
            sliceafter = [iter_all, iter_after]

        # if the length of the scan is a fraction (likely), the final pixel will only have a fraction of a
        # pixel's worth of flux.
        def subpixelcheck(i, val):
            if val - i > 1:
                return 1
            else:
                return val - i

        for i in range(int(np.ceil(scan_length_pix))):
            # Make sure there is a scan angle before doing the complex shift() calculation
            if scan_angle != 0.0:
                # TODO for imaging scan: this is a shift WITHIN an array; we want a subpixel translation OF
                # the array, ideally
                # TODO: Delete this if WFC3 team does not require.
                scan_product[slicelist[0][i],slicelist[1][i]] += shift(
                    product * subpixelcheck(i,scan_length_pix),
                    shift=(np.cos(scan_angle)*dpix[i], np.sin(scan_angle)*dpix[i]),
                    mode='nearest',
                    order=1
                ) / scan_length_pix
            else:
                scan_product[slicelist[0][i],slicelist[1][i]] += product * subpixelcheck(i,scan_length_pix) / scan_length_pix

            # Adding background to all other pixels, unless we are asked not to.
            if add_extended_background:
                scan_product[slicebefore[0][i],slicebefore[1][i]] += bgvals * subpixelcheck(i,scan_length_pix) / scan_length_pix
                scan_product[sliceafter[0][i], sliceafter[1][i]] += bgvals * subpixelcheck(i,scan_length_pix) / scan_length_pix
        return scan_product

    def slitless_scan_rate(self, rate, add_extended_background=True):
        '''
        Calculate a slitless scan projection.

        How this works: 
        1. We calculate a regular slitless projection
        2. We recompute the necessary background values that don't get passed out of slitless_rate
        3. We run the scan-projection code on each of the three products of the slitless projection
        4. We recompute the binning object

        Parameters
        ----------
        rate: 3D numpy.ndarray
            Cube containing the flux rate at the focal plane
        add_extended_background: bool (default: True)
            Toggle for including extended background not contained within the flux cube

        Returns
        -------
        products: 4 entry tuple
            wave_pix: 1D numpy.ndarray containing wavelength to pixel mapping on the detector plane
            spec_rate: 2D numpy.ndarray of detector count rates
            spec_no_qyield: 2D numpy.ndarray of the detector rate without quantum yield correction
            spec_variance: 2D numpy.ndarray of the variance array.
        '''
        wave_pix, fp_pix_rate, fp_pix_no_qyield, fp_pix_variance = self.slitless_rate(
                rate,
                add_extended_background=add_extended_background
            )

        
        # Interpolate the background onto the output/instrument pixel spacing
        int_bg_fp_rate = sci_int.interp1d(self.wave, self.bg_fp_rate.astype(np.float32, casting='same_kind'),
                                          kind='linear', assume_sorted=True, copy=False)
        bg_fp_rate_pix = int_bg_fp_rate(wave_pix)
        q_yield, fano_factor = self.current_instrument.get_quantum_yield(wave_pix)
        # To meet IDT expectations, some instruments require a possibly chromatic fudge factor to be applied
        # To the per-pixel electron rate variance.
        var_fudge = self.current_instrument.get_variance_fudge(wave_pix)
        # Calculate electron rate and variance due to background (per wavelength -> scalar)
        bg_electron_rate = integrate.trapz(bg_fp_rate_pix * q_yield, wave_pix)
        bg_rate_no_qyield = integrate.trapz(bg_fp_rate_pix, wave_pix)
        bg_electron_variance = integrate.trapz(bg_fp_rate_pix * q_yield * (q_yield + fano_factor) * var_fudge, wave_pix)

        # Scan the projection.
        fp_pix_rate = self.observation_scan(fp_pix_rate, bg_electron_rate, add_extended_background=add_extended_background)
        fp_pix_no_qyield = self.observation_scan(fp_pix_no_qyield, bg_rate_no_qyield, add_extended_background=add_extended_background)
        fp_pix_variance = self.observation_scan(fp_pix_variance, bg_electron_variance, add_extended_background=add_extended_background)

        # Expand the grid so we can extract a box that tall.
        # That means extending nx or ny to the new cross-dispersion height.
        # This must be done before binning, because binning rewrites the grid into an
        # IrregularGrid.
        if self.dispersion_axis == "x":
            ny = fp_pix_rate.shape[0]
            nx = self.grid.nx
        else:
            ny = self.grid.ny
            nx = fp_pix_rate.shape[1]
        newgrid = coords.Grid(self.grid.xsamp,self.grid.ysamp, nx, ny)
        self.grid = newgrid

        # bin the instrument wavelength grid in pixels
        if self.dispersion_axis == 'x':
            bin_x = self.current_instrument.get_dispersion_binning()
            bin_y = self.current_instrument.get_spatial_binning()
        else:
            bin_x = self.current_instrument.get_spatial_binning()
            bin_y = self.current_instrument.get_dispersion_binning()
        self.binning = DetectorBinning(fp_pix_rate, bin_x, bin_y, self.grid, dispersion_axis=self.dispersion_axis, wave=wave_pix, projection_type=self.projection_type)

        # slitless_rate has already taken care of determining if we need to flip the projected image upside down
        products = wave_pix, fp_pix_rate, fp_pix_no_qyield, fp_pix_variance

        return products

    def image_scan_rate(self, rate, add_extended_background=True):
        '''
        Calculate an imaging scan projection.

        How this works:
        1. We calculate a regular imaging projection
        2. We recompute the necessary background values that don't get passed out of image_rate
        3. We run the scan-projection code on each of the three products of the image projection
        4. We recompute the binning object

        Parameters
        ----------
        rate: 3D numpy.ndarray
            Cube containing the flux rate at the focal plane
        add_extended_background: bool (default: True)
            Toggle for including extended background not contained within the flux cube

        Returns
        -------
        products: 4 entry tuple
            wave_pix: 1D numpy.ndarray containing wavelength to pixel mapping on the detector plane
            fp_pix_rate: 2D numpy.ndarray of detector count rates
            fp_pix_no_qyield: 2D numpy.ndarray of the detector rate without quantum yield correction
            fp_pix_variance: 2D numpy.ndarray of the variance array.
        '''
        fp_pix_rate, fp_pix_no_qyield, fp_pix_variance = self.image_rate(rate)
        wave_pix = self.wave_eff(rate)
        
        # Interpolate the background onto the output/instrument pixel spacing
        int_bg_fp_rate = sci_int.interp1d(self.wave, self.bg_fp_rate.astype(np.float32, casting='same_kind'),
                                          kind='linear', assume_sorted=True, copy=False)
        bg_fp_rate_pix = int_bg_fp_rate(self.wave)

        q_yield, fano_factor = self.current_instrument.get_quantum_yield(self.wave)
        # To meet IDT expectations, some instruments require a possibly chromatic fudge factor to be applied
        # To the per-pixel electron rate variance.
        var_fudge = self.current_instrument.get_variance_fudge(self.wave)
        # Calculate electron rate and variance due to background (per wavelength -> scalar)
        bg_electron_rate = integrate.trapz(bg_fp_rate_pix * q_yield, self.wave)
        bg_rate_no_qyield = integrate.trapz(bg_fp_rate_pix, self.wave)
        bg_electron_variance = integrate.trapz(bg_fp_rate_pix * q_yield * (q_yield + fano_factor) * var_fudge, self.wave)

        # Scan the projection.
        fp_pix_rate = self.observation_scan(fp_pix_rate, bg_electron_rate, add_extended_background=add_extended_background)
        fp_pix_no_qyield = self.observation_scan(fp_pix_no_qyield, bg_rate_no_qyield, add_extended_background=add_extended_background)
        fp_pix_variance = self.observation_scan(fp_pix_variance, bg_electron_variance, add_extended_background=add_extended_background)
 
        # Expand the grid so we can extract a box that tall.
        # That means extending nx or ny to the new cross-dispersion height.
        # This must be done before binning, because binning rewrites the grid into an
        # IrregularGrid.
        ny = fp_pix_rate.shape[0]
        nx = self.grid.nx
        newgrid = coords.Grid(self.grid.xsamp,self.grid.ysamp, nx, ny)
        self.grid = newgrid

        # bin the instrument wavelength grid in pixels
        bin_x = self.current_instrument.get_spatial_binning()
        bin_y = self.current_instrument.get_spatial_binning()
        self.binning = DetectorBinning(fp_pix_rate, bin_x, bin_y, self.grid)

        products = wave_pix, fp_pix_rate, fp_pix_no_qyield, fp_pix_variance

        return products

    def wave_eff(self, rate):
        if len(rate.shape) > 1:
            rate_tot = np.nansum(rate, axis=0)
        else:
            rate_tot = rate
        a = np.sum(rate_tot * self.wave)
        b = np.sum(rate_tot)

        if (b > 0.0) and (a > 0.0):
            wave_eff = a / b
        else:
            wave_eff = self.wave.mean()
        wave_eff_arr = np.array([wave_eff])
        return wave_eff_arr

    def get_projection_type(self):
        return self.projection_type

    def ipc_convolve(self, rate, kernel):
        fp_pix_ipc = sg.fftconvolve(rate, kernel, mode='same')

        debug_utils.debugarrays.store('signal', 'ipc_convolve',
                                      {
                                          'rate': rate,
                                          'kernel': kernel,
                                          'fp_pix_ipc': fp_pix_ipc,
                                          'description': 'This is just a short, unnecessary description.'
                                      })

        return fp_pix_ipc

    def get_saturation_mask(self, rate=None):
        """
        Return a numpy array indicating pixels with full saturation (2), partial saturation (1) and no saturation (0).

        Parameters
        ----------
        rate: None or 2D np.ndarray
            Detector plane rate image used to build saturation map from

        Returns
        -------
        mask: 2D np.ndarray
            Saturation mask image
        """
        if rate is None:
            rate = self.rate_plus_bg_unbinned

        # Fill the saturation mask using the _get_saturation_mask method
        saturation_mask = self.the_detector._get_saturation_mask(rate)

        saturation_mask = self.binning.max(saturation_mask)

        return np.asarray(saturation_mask, dtype=np.int8)

    def calc_total_detector(self, rate, bg_pix_rate):
        """ 
        Compute the total detector rate.

        Pandeia only handles a small "postage stamp" FOV, which means any given
        DetectorSignal only covers a portion of the entire detector.

        The total detector rate is built out of four rates:
        1. The source signal (from rate_no_qyield). Should be entirely 
           contained in that computed 2D projection.
        2. The sky (and thermal) background flux. If there's a slit, it should be
           contained entirely within the slit height and dispersed width, and it has 
           wavelength dependence in the dispersion direction. If there's no slit, it 
           covers the entire exposed portion of the detector and is gray.
        3. The dark current. Covers the entire detector, exposed or not.
        4. Scattered light, if defined. Covers the same region as the sky background.

        Other assumptions:
        The 2D projection may be a focal plane consisting of more than one detector. 
        In this case, the dark_current is not a scalar, but a 2D array giving the
        spatially variable dark current that will have to be cropped to the single
        detector.

        The background we have is a background that CONTAINS the dark current, which 
        must be subtracted off before further processing. We then have a clean background
        over the area of the "postage stamp" which we can scale accordingly. In the case 
        of a slit, even though the background is not gray it's only being extended in one
        direction, so we can still simply multiply the background flux by a scaling factor.
        The outlying case is a slit whose height is less than the postage stamp, in which
        case it's fully contained by the bg_pix_rate array and no scaling is needed.

        """
        chips = self.current_instrument.get_chip_dimensions(self.wave_pix)

        detector_total_rate_dict = {}
        for chip in chips:
            ## Process the source signal
            rate_temp = rate[chip["bounds"]["y"],chip["bounds"]["x"]]
            rate_full = np.sum(rate_temp)

            ## Process the dark current
            # dark current is either a scalar (float) or 2D map
            if isinstance(self.the_detector.dark_current, np.ndarray):
                dc = self.the_detector.dark_current[chip["bounds"]["y"],chip["bounds"]["x"]]
                dc_full = np.sum(dc)
            else:
                dc = self.the_detector.dark_current
                dc_full = chip["size"]["y"] * chip["size"]["x"] * dc
                
            ## Process the background
            bg = bg_pix_rate[chip["bounds"]["y"],chip["bounds"]["x"]] 
            bg -= np.mean(dc) # dc is either a scalar or the entire size of the detector
            if self.projection_type == "spec":
                if self.dispersion_axis == "x":
                    bgx = rate_temp.shape[1]
                    # if this is fractional, it's actually fine. All we need is the number, anyway.
                    bgy = self.current_instrument.get_slit_pars()['xdisp'] / \
                               self.current_instrument.get_aperture_pars()["plate_scale"][0]
                    
                else:
                    bgx = self.current_instrument.get_slit_pars()['xdisp'] / \
                               self.current_instrument.get_aperture_pars()["plate_scale"][1]
                    bgy = rate_temp.shape[0]
                
            else:
                bgx = chip["size"]["x"]
                bgy = chip["size"]["y"]
            # if either slit-limited dimension is SMALLER than the background array, the background
            # array contains the entire background flux and needs no scaling.
            if bgx < chip["size"]["x"] or bgy < chip["size"]["y"]:
                bg_full = np.sum(bg)
            else:
                # we are multiplying the background by a factor of how much background is covered by the slit
                bg_full = np.mean(bg) * bgy * bgx

            total = rate_full + dc_full + bg_full 

            # selection logic: If calculation_config specifies it, do that. If not, use the predefined data value
            scatter = scatter = self.calculation_config.noise['scatter'] if self.calculation_config.noise['scatter'] is not None else self.the_detector.scatter
            if scatter:
                ## Process scattered light
                # Scattering depends on the disperser and cenwave
                global_scattering = self.current_instrument.get_global_scattering()
                scattered_full = global_scattering * rate_full

                total += scattered_full

            detector_total_rate_dict[chip["name"]] = total

        return detector_total_rate_dict

    def calc_brightest_pixel(self, rate, bg_pix_rate):
        """
        Compute the brightest pixel rate.

        Pandeia only handles a small "postage stamp" FOV, which means any given
        DetectorSignal only covers a portion of the entire detector. Using the assumption
        that all the source flux is contained within that postage stamp, we can extend the
        background and dark current across the entire detector.

        This assumption is only valid if the scene has a single, centered point source.
        Any other configuration should trigger a warning.

        """
        # define 2x2 kernel
        kernel = np.asarray([[0.25,0.25],[0.25,0.25]], dtype=np.float32)

        chips = self.current_instrument.get_chip_dimensions(self.wave_pix)

        brightest_pixel_rate_dict = {}
        for chip in chips:
            temp_rate = copy.deepcopy(rate[chip["bounds"]["y"],chip["bounds"]["x"]])

            # bg_pix_rate contains bg, postflash, and dc rates. They SHOULD ultimately
            # both be single-valued
            bg = np.mean(bg_pix_rate[chip["bounds"]["y"],chip["bounds"]["x"]])

            # a chip may exist but be devoid of data. Such as with certain combinations of
            # grating and central wavelength in COS.
            if np.isnan(bg):
                break

            temp_rate += bg

            # measure the location of the brightest pixel (in wavelength, only meaningful
            # for spectral projections)
            bploc = np.unravel_index(np.argmax(temp_rate), temp_rate.shape)
            if self.projection_type == "spec":
                if self.dispersion_axis == "x":
                    wave_pix_temp = copy.deepcopy(self.wave_pix[chip["bounds"]["x"]])
                    bp_wave = wave_pix_temp[bploc[1]]
                else:
                    wave_pix_temp = copy.deepcopy(self.wave_pix[chip["bounds"]["y"]])
                    bp_wave = wave_pix_temp[bploc[0]]
                bp_val = temp_rate[bploc]
                    
            elif self.projection_type in ("image", "image_scan"):
                # This is step 5 of the process; a pixel-phase-agnostic
                # brightest_pixel_rate_imaging was created by
                # ConvolvedSceneCube.brightest_pixel_rate()
                bp_wave = self.wave_pix[0]
                bp_val = self.brightest_pixel_rate_imaging + bg
            elif self.projection_type in ("slitless", "slitless_scan", "multiorder"):
                # The idea of a wavelength-of-brightest-pixel is problematic for slitless
                # and multiorder modes. It can only be reasonably defined for a single
                # centered single-order source. It fails when a.) the target source is not
                # centered, b.) the brightest pixel isn't IN the target source (it won't
                # crash if that brightest pixel is still within the bounds of a centered
                # source, but it'll be wrong), or c.) when the mode uses CombinedSignal
                # and there's more than one wavelength grid superimposed on the image. The
                # values produced by this function are not used in JWST or Roman results
                # The cross-dispersion position should match the source that has the
                # bright pixel in it Then, use the dispersion position to find the
                # associated wavelength.
                if not self.single_centered_source:
                    key = "brightest_pixel_wavelength_centered"
                    self.warnings[key] = warning_messages[key]
                if self.dispersion_axis == "x":
                    dispersion_dimension=1
                else:
                    dispersion_dimension=0
                # wave_pix_temp will be smaller than the corresponding dimension of the 2D
                # image
                wave_pix_temp = copy.deepcopy(self.wave_pix[chip["bounds"][self.dispersion_axis]])
                excess = temp_rate.shape[dispersion_dimension] - len(wave_pix_temp)
                # it's nans, only filled in at the middle where the default wavelength
                # grid is defined.
                wave_pix_temp2 = np.zeros(temp_rate.shape[dispersion_dimension]) * np.nan
                wave_pix_temp2[excess//2:excess//2+len(wave_pix_temp)] = wave_pix_temp
                bp_wave = wave_pix_temp2[bploc[dispersion_dimension]]
                bp_val = temp_rate[bploc]
            else:
                raise EngineInputError("Unrecognized projection type")
            brightest_pixel_rate_dict[chip["name"]] = [bp_val, bp_wave]

        return brightest_pixel_rate_dict

class DetectorBinning(object):
    def __init__(self, rate, bin_x, bin_y, oldgrid, dispersion_axis="x", wave=None, projection_type=""):
        """
        This class exists to model on-detector binning, where individual pixels are exposed and then combined
        with their neighbors within the readout electronics. It can handle arbitrary positive integer binning
        values.

        When instantiated, the class computes the parameters needed to crop and rebin data, as well as the new
        coordinate IrregularGrid. Binning can then be applied to any 2D detector product that uses the same
        coordinate grid by running one of the methods. Because of the nature of the binning, multiple methods
        are offered. For instance: We wish to conserve flux within a binned image, so we use the sum() method
        to sum the pixels. However, saturation handling must be different: we can no longer use the fullwell
        value to determine the binning; the fundamental thing that is saturating is the original pixel
        electronics, and thus a 2x2 binned pixel with 3x the fullwell flux might be constituted of four pixels
        that each reached 75% fullwell (in which case the flux is not saturated), or it could be a single
        highly saturated pixel (in which case the flux for that binned pixel is unreliable). Accordingly, we
        need to compute saturation on the unbinned array, and then set each bin to the max() of the saturation
        of the constituent pixels.

        The basic premise is that binning a 2D image requires three operations:
        1. Crop the array to an even multiple of the binning value
        2. Bin the cropped array
        3. Rewrite the spatial grid so it matches the new cropped, binned dimensions

        Binning a wavelength array requires four operations:
        1. Crop the array to an even multiple of the binning value
        2. Bin the cropped array
        3. Rewrite the spatial grid so it matches the new cropped, binned dimensions
        4. Rewrite the wavelength array using the mean of the values of the binned pixels
        With the added stipulation that for slitless spectroscopy, we are concerned with making the slitless
        spectroscopy trace completely fill the first binned pixel (so that we don't have to crop it later)
        
        The geometry of the slitless mode is basically that of the imaging scene with an expanded section in
        the middle corresponding to the wavelength array: 
        [   *   ] -> [   --------   ]
        We need to make sure that middle section lines up so that the slitless spectrum starts on a pixel: we
        lose at most one partially-covered pixel at the high-wavelength end that way.

        Methods:
        --------
        sum:
            Rebins and sums the pixels that form each bin. This conserves flux for signal products
        mean:
            Rebins and takes the mean of the pixels that form each bin. This conserves mask shape for
            extraction masks
        max:
            Rebins and takes the max of the pixels that form each bin (for the saturation mask)
        min:
            Rebins and takes the min of the pixels that form each bin (for the groups-before-sat image)
        dispersion: 
            Rebins the wavelength array, taking the mean of the wavelengths that form each binned pixel.
        """
        # your basic binning needs to be cropped to the nearest multiple of the binning
        self.cropshape = [int(np.floor(rate.shape[0]/bin_y) * bin_y), int(np.floor(rate.shape[1]/bin_x) * bin_x)]
        # and then, of course, binned.
        self.newshape = [int(self.cropshape[0]/bin_y), int(self.cropshape[1]/bin_x)]

        # by default we will assume we want to center the result - trim evenly off the edges
        # y is backwards
        self.end_y = int(np.floor((rate.shape[0]-self.cropshape[0])/2))
        self.start_x = int(np.floor((rate.shape[1]-self.cropshape[1])/2))
        # Now it gets complicated: what kind of rebinning are we doing? image, slit, or slitless?
        if wave is not None:
            if dispersion_axis == 'x':
                bin_wave = bin_x
                self.cropwaveshape = int(np.floor(wave.shape[0]/bin_wave) * bin_wave)
                self.newwaveshape = int(np.floor(wave.shape[0]/bin_wave))
                self.start_wave = self.start_x
                if projection_type in ("slitless", "slitless_scan"):
                    # This next equation finds how many pixels there are before the 
                    # wavelength array starts. If that's an even multiple of the binning
                    # we need to do nothing. 
                    self.start_wave = 0
                    offset = int((self.cropshape[1]-len(wave))/2 % bin_wave)
                    if offset % bin_wave > 0:
                        self.start_x += offset
                        # startx was already set to the best position; moving it
                        # most likely means losing another pixel from the far side
                        self.cropshape[1] -= bin_wave
                        self.newshape[1] -= 1 
            else:
                bin_wave = bin_y
                self.cropwaveshape = int(np.floor(wave.shape[0]/bin_wave) * bin_wave)
                self.newwaveshape = int(np.floor(wave.shape[0]/bin_wave))
                self.start_wave = self.end_y
                if projection_type in ("slitless", "slitless_scan"):
                    self.start_wave = 0
                    # we want to line up the low-wavelength end, which is at the high-index end
                    offset = bin_wave - int((self.cropshape[0]+len(wave))/2 % bin_wave)
                    if offset % bin_wave > 0:
                        self.end_y += offset
                        self.cropshape[0] -= bin_wave
                        self.newshape[0] -= 1 

        # now, rewrite the Grid instance as an IrregularGrid.
        # Find the actual min x and y position (relating to the edges of the pixels)
        gridy = oldgrid.col[self.end_y: self.end_y+self.cropshape[0]]
        gridx = oldgrid.row[self.start_x: self.start_x+self.cropshape[1]]
        start_y = np.min(gridy) - 0.5 * oldgrid.ysamp
        start_x = np.min(gridx) - 0.5 * oldgrid.xsamp

        # And the new y sampling and number of pixels
        newsampy = bin_y * oldgrid.ysamp
        newsampx = bin_x * oldgrid.xsamp
        # create a new array that starts in the middle of the first pixel and goes to the middle of the last pixel
        gridyvals = np.linspace(start_y + 0.5 * newsampy, start_y + newsampy * (self.newshape[0] - 0.5), self.newshape[0])[::-1]
        gridxvals = np.linspace(start_x + 0.5 * newsampx, start_x + newsampx * (self.newshape[1] - 0.5), self.newshape[1])

        self.grid = coords.IrregularGrid(gridyvals, gridxvals)

    def sum(self, rate):
        # this MUST be sliced in numpy notation ([x,y]) rather than sequentially [x][y]) 
        croprate = rate[self.end_y:self.end_y+self.cropshape[0], self.start_x:self.start_x+self.cropshape[1]]

        # rebin the 2D array
        sh = int(self.newshape[0]), int(self.cropshape[0] // self.newshape[0]), int(self.newshape[1]), int(self.cropshape[1] // self.newshape[1])
        return croprate.reshape(sh).sum(-1).sum(1)

    def mean(self, rate):
        # this MUST be sliced in numpy notation ([x,y]) rather than sequentially [x][y]) 
        croprate = rate[self.end_y:self.end_y+self.cropshape[0], self.start_x:self.start_x+self.cropshape[1]]

        # rebin the 2D array
        sh = int(self.newshape[0]), int(self.cropshape[0] // self.newshape[0]), int(self.newshape[1]), int(self.cropshape[1] // self.newshape[1])
        return croprate.reshape(sh).mean(-1).mean(1)

    def min(self, rate):
        # this MUST be sliced in numpy notation ([x,y]) rather than sequentially [x][y]) 
        croprate = rate[self.end_y:self.end_y+self.cropshape[0], self.start_x:self.start_x+self.cropshape[1]]

        # rebin the 2D array
        sh = int(self.newshape[0]), int(self.cropshape[0] // self.newshape[0]), int(self.newshape[1]), int(self.cropshape[1] // self.newshape[1])
        return croprate.reshape(sh).min(-1).min(1)

    def max(self, rate):
        # this MUST be sliced in numpy notation ([x,y]) rather than sequentially [x][y]) 
        croprate = rate[self.end_y:self.end_y+self.cropshape[0], self.start_x:self.start_x+self.cropshape[1]]

        # rebin the 2D array
        sh = int(self.newshape[0]), int(self.cropshape[0] // self.newshape[0]), int(self.newshape[1]), int(self.cropshape[1] // self.newshape[1])
        return croprate.reshape(sh).max(-1).max(1)

    def dispersion(self, wave):
        if hasattr(self, "start_wave"):
            # We have now set up the cropping required.
            cropwave = wave[self.start_wave:self.start_wave+self.cropwaveshape]
            # rebin the wavelength array
            wavesh = int(self.newwaveshape), int(self.cropwaveshape // self.newwaveshape)
            # we take the mean rather than the sum, because we want the new midpoint
            # wavelength arrays generally aren't linear, so some info is lost here.
            return cropwave.reshape(wavesh).mean(-1)
        else:
            return wave

class CombinedSignal(object):
    """
    This class takes a set of DetectorSignal instances, combines the rates appropriately, and
    provides the other information that DetectorNoise requires.  The primary use for this is in the case
    of SOSS where the detector plane contains signals from effectively three instrument configurations, one
    for each of the visible orders of the gr700xd disperser. These signals need to be combined properly before
    being used to calculate a DetectorNoise and perform a Strategy extraction.

    WARNING: This is currently set up to only support a single aperture slice. Supporting multiple orders with
    multiple slices will require further refactoring...

    Parameters
    ----------
    signal_list: list
        list of DetectorSignal instances
    """

    def __init__(self, signal_list):
        # each signal potentially has a different grid so we want to combine everything onto one grid. instruments
        # that combine multiple signals need to provide detector_pixels so that we know how to shift the signals
        # with respect to each other.
        self.warnings = {}
        self.detector_pixel_list = []
        self.wave_pix_dict = {}
        # some things are common to all signals so get them from the first one
        self.parent_signal = signal_list[0]
        self.signal_list = signal_list
        self.dispersion_axis = self.parent_signal.dispersion_axis
        self.binning = self.parent_signal.binning
        self.validate = self.parent_signal.validate
        self.webapp = self.parent_signal.webapp

        if len(self.parent_signal.rate_list) > 1:
            msg = "Combining multiple instrument signals onto single detector only supports a single slice."
            raise NotImplementedError(value=msg)

        maxx = 0
        maxy = 0
        maxx_unbinned = 0
        maxy_unbinned = 0
        for i, s in enumerate(self.signal_list):
            self.warnings.update(s.warnings)
            pixels = s.detector_pixels
            if pixels is None:
                inst = s.current_instrument.instrument['instrument']
                msg = "No detector pixel configuration set for instrument %s." % inst
                raise DataError(value=msg)
            self.detector_pixel_list.append(pixels)
            ny, nx = s.rate.shape
            maxx = max(maxx, nx)
            maxy = max(maxy, ny)
            ny, nx = s.rate_plus_bg_unbinned.shape
            maxx_unbinned = max(maxx_unbinned, nx)
            maxy_unbinned = max(maxy_unbinned, ny)
            self.wave_pix_dict[s.order] = s.wave_pix

        if self.parent_signal.projection_type == "multiorder":
            self.grid = coords.IrregularGrid(np.arange(maxy), np.arange(maxx) - maxx)
            self.grid_unbinned = coords.IrregularGrid(np.arange(maxy_unbinned), np.arange(maxx_unbinned) - maxx_unbinned)
        else:
            if self.dispersion_axis == 'x':
                # for slitless calculations, the dispersion axis is longer than the spectrum being dispersed
                # because the whole field of view is being dispersed. 'excess' is the size of the FOV
                # and half will be to the left of the blue end of the spectrum and half to the right of the red end.
                # this is used to create the new spatial coordinate transform for the pixel image on the detector.
                self.grid = coords.IrregularGrid(
                    self.parent_signal.grid.col,
                    (np.arange(maxx) - (maxx) / 2.0) * self.parent_signal.grid.xsamp
                )
                self.grid_unbinned = coords.IrregularGrid(
                    self.parent_signal.grid_unbinned.col,
                    (np.arange(maxx_unbinned) - (maxx_unbinned) / 2.0) * self.parent_signal.grid_unbinned.xsamp
                )
            else:
                self.grid = coords.IrregularGrid(
                    (np.arange(maxy) - (maxy) / 2.0) * self.parent_signal.grid.ysamp,
                    self.parent_signal.grid.row
                )
                self.grid_unbinned = coords.IrregularGrid(
                    (np.arange(maxy_unbinned) - (maxy_unbinned) / 2.0) * self.parent_signal.grid_unbinned.ysamp,
                    self.parent_signal.grid_unbinned.row
                )

        self.pixgrid_list = [self.grid]
        if self.validate:
            self.dist = self.grid.dist()
            self.dist_unbinned = self.grid_unbinned.dist()
        self.det_mask = np.ones((maxy,maxx))

        # these parameters are the same for each signal
        self.wave = self.parent_signal.wave
        self.total_flux = self.parent_signal.total_flux
        self.total_rate = self.parent_signal.total_rate
        self.bg_fp_rate = self.parent_signal.bg_fp_rate
        self.background = self.parent_signal.background
        self.flux_cube_list = self.parent_signal.flux_cube_list
        self.flux_plus_bg_list = self.parent_signal.flux_plus_bg_list
        if self.validate:
            self.fp_rate = self.parent_signal.fp_rate


        self.aperture_list = self.parent_signal.aperture_list
        self.projection_type = self.parent_signal.projection_type
        self.current_instrument = self.parent_signal.current_instrument
        self.spatial_grid = self.parent_signal.grid
        self.spatial_grid_unbinned = self.parent_signal.grid_unbinned
        self.the_detector = self.parent_signal.the_detector
        self.calculation_config = self.parent_signal.calculation_config
        if self.parent_signal.the_detector.rn_correlation:
            self.read_noise_correlation_matrix = self.parent_signal.read_noise_correlation_matrix

        self.unbinned_array = np.zeros((maxy_unbinned,maxx_unbinned))

        self.rate_list = [{
            'fp_pix': np.zeros((maxy,maxx)),
            'fp_pix_no_ipc': np.zeros((maxy,maxx)),
            'fp_pix_no_qyield': np.zeros((maxy,maxx)),
            'fp_pix_variance': np.zeros((maxy,maxx)),
            'fp_pix_unbinned': np.zeros((maxy_unbinned,maxx_unbinned)),
            'fp_pix_no_ipc_unbinned': np.zeros((maxy_unbinned,maxx_unbinned)),
            'fp_pix_no_qyield_unbinned': np.zeros((maxy_unbinned,maxx_unbinned)),
            'fp_pix_variance_unbinned': np.zeros((maxy_unbinned,maxx_unbinned))
        }]
        self.rate_plus_bg_list = [{
            'fp_pix': np.zeros((maxy,maxx)),
            'fp_pix_no_ipc': np.zeros((maxy,maxx)),
            'fp_pix_no_qyield': np.zeros((maxy,maxx)),
            'fp_pix_variance': np.zeros((maxy,maxx)),
            'fp_pix_unbinned': np.zeros((maxy_unbinned,maxx_unbinned)),
            'fp_pix_no_ipc_unbinned': np.zeros((maxy_unbinned,maxx_unbinned)),
            'fp_pix_no_qyield_unbinned': np.zeros((maxy_unbinned,maxx_unbinned)),
            'fp_pix_variance_unbinned': np.zeros((maxy_unbinned,maxx_unbinned)),
        }]
        saturation = np.zeros((maxy_unbinned,maxx_unbinned))

        for i, s in enumerate(self.signal_list):
            # each fp_pix and fp_pix_no_ipc product already has background and dark current (and possibly
            # other things) added. to avoid double-counting the additional flux sources, we need to subtract
            # them from each piece and add it back in once at the end.
            s.rate_plus_bg_list[0]['fp_pix_unbinned'] -= s.additional_flux
            s.rate_plus_bg_list[0]['fp_pix_no_ipc_unbinned'] -= s.additional_flux
            s.rate_plus_bg_list[0]['fp_pix_variance_unbinned'] -= s.additional_flux

            ny, nx = s.rate_list[0]['fp_pix_unbinned'].shape
            # set up position within the combined rate image to add each rate
            if self.dispersion_axis == 'x':
                ly = 0
                uy = int(self.unbinned_array.shape[0] - s.rate_list[0]['fp_pix_unbinned'].shape[0])
                # shifts between the rates are referenced to the first pixel location
                lx = 0
                ux = int(self.unbinned_array.shape[1] - s.rate_list[0]['fp_pix_unbinned'].shape[1])
            else:
                # shifts between the rates are referenced to the first pixel location
                ly = 0
                uy = int(self.unbinned_array.shape[0] - s.rate_list[0]['fp_pix_unbinned'].shape[0])
                lx = 0
                ux = int(self.unbinned_array.shape[1] - s.rate_list[0]['fp_pix_unbinned'].shape[1])

            # this will cause pixels that don't overlap to be filled with the background. this is correct for imaging
            # and slitless, but may need revisiting for other cases.
            new_r = np.pad(s.rate_list[0]['fp_pix_unbinned'], ([ly, uy], [lx, ux]), mode='edge')
            new_r_bg = np.pad(s.rate_plus_bg_list[0]['fp_pix_unbinned'], ([ly, uy], [lx, ux]), mode='edge')
            new_r_noipc = np.pad(s.rate_list[0]['fp_pix_no_ipc_unbinned'], ([ly, uy], [lx, ux]), mode='edge')
            new_r_bg_noipc = np.pad(s.rate_plus_bg_list[0]['fp_pix_no_ipc_unbinned'], ([ly, uy], [lx, ux]), mode='edge')
            new_r_noqyield = np.pad(s.rate_list[0]['fp_pix_no_qyield_unbinned'], ([ly, uy], [lx, ux]), mode='edge')
            new_r_bg_noqyield = np.pad(s.rate_plus_bg_list[0]['fp_pix_no_qyield_unbinned'], ([ly, uy], [lx, ux]), mode='edge')
            new_r_var = np.pad(s.rate_list[0]['fp_pix_variance_unbinned'], ([ly, uy], [lx, ux]), mode='edge')
            new_r_bg_var = np.pad(s.rate_plus_bg_list[0]['fp_pix_variance_unbinned'], ([ly, uy], [lx, ux]), mode='edge')
            saturation_mask = self.the_detector._get_saturation_mask(s.rate_plus_bg_list[0]['fp_pix_no_ipc_unbinned'])
            new_sat = np.pad(saturation_mask, ([ly, uy], [lx, ux]), mode='constant')

            saturation = np.maximum(saturation, new_sat)

            self.rate_list[0]['fp_pix_unbinned'] += new_r
            self.rate_list[0]['fp_pix_no_ipc_unbinned'] += new_r_noipc
            self.rate_list[0]['fp_pix_no_qyield_unbinned'] += new_r_noqyield
            self.rate_list[0]['fp_pix_variance_unbinned'] += new_r_var
            self.rate_plus_bg_list[0]['fp_pix_unbinned'] += new_r_bg
            self.rate_plus_bg_list[0]['fp_pix_no_ipc_unbinned'] += new_r_bg_noipc
            self.rate_plus_bg_list[0]['fp_pix_no_qyield_unbinned'] += new_r_bg_noqyield
            self.rate_plus_bg_list[0]['fp_pix_variance_unbinned'] += new_r_bg_var

        # now we add the dark current back in
        # this assumes a scalar dark level; there is no allowance for a non-uniform dark to be added
        # back here.
        self.rate_plus_bg_list[0]['fp_pix_unbinned'] += self.parent_signal.additional_flux
        self.rate_plus_bg_list[0]['fp_pix_no_ipc_unbinned'] += self.parent_signal.additional_flux
        self.rate_plus_bg_list[0]['fp_pix_variance_unbinned'] += self.parent_signal.additional_flux

        self.rate_list[0]['fp_pix'] = self.binning.sum(self.rate_list[0]['fp_pix_unbinned'])
        self.rate_list[0]['fp_pix_no_ipc'] = self.binning.sum(self.rate_list[0]['fp_pix_no_ipc_unbinned'])
        self.rate_list[0]['fp_pix_no_qyield'] = self.binning.sum(self.rate_list[0]['fp_pix_no_qyield_unbinned'])
        self.rate_list[0]['fp_pix_variance'] = self.binning.sum(self.rate_list[0]['fp_pix_variance_unbinned'])
        self.rate_plus_bg_list[0]['fp_pix'] = self.binning.sum(self.rate_plus_bg_list[0]['fp_pix_unbinned'])
        self.rate_plus_bg_list[0]['fp_pix_no_ipc'] = self.binning.sum(self.rate_plus_bg_list[0]['fp_pix_no_ipc_unbinned'])
        self.rate_plus_bg_list[0]['fp_pix_no_qyield'] = self.binning.sum(self.rate_plus_bg_list[0]['fp_pix_no_qyield_unbinned'])
        self.rate_plus_bg_list[0]['fp_pix_variance'] = self.binning.sum(self.rate_plus_bg_list[0]['fp_pix_variance_unbinned'])

        self.saturation_list = [self.binning.max(saturation)]
        # SOSS only has one aperture so the on_detector rates are just the fp_pix rates.
        self.rate = self.rate_list[0]['fp_pix']
        self.rate_no_qyield = self.rate_list[0]['fp_pix_no_qyield']
        self.rate_plus_bg = self.rate_plus_bg_list[0]['fp_pix']
        self.rate_unbinned = self.rate_list[0]['fp_pix_unbinned']
        self.rate_plus_bg_unbinned = self.rate_plus_bg_list[0]['fp_pix_unbinned']

        # this code must be kept in sync with similar code in DetectorSignal.signal_products()
        self.bg_pix_rate = self.rate_plus_bg_unbinned - self.rate_unbinned
        # Check to see if the background is saturating (on unbinned data)
        bgsat = self.get_saturation_mask(self.bg_pix_rate)
        if (np.sum(bgsat) > 0) or (np.isnan(np.sum(bgsat))):
            key = "background_saturated"
            self.warnings[key] = warning_messages[key]
        # bin to match the rest of the data
        self.bg_pix_rate = self.binning.sum(self.bg_pix_rate)

        # create ngroup map from the newly combined signal
        self.ngroup_map = self.signal_list[0].the_detector.get_before_sat(self.rate_plus_bg_unbinned)
        self.ngroup_map = self.binning.min(self.ngroup_map)
        self.fraction_saturation = np.max(
            self.signal_list[0].the_detector.get_saturation_fraction(self.rate_plus_bg_unbinned))
        self.brightest_pixel = np.max(self.rate_plus_bg)
        self.groups_list = [self.ngroup_map]

        if isinstance(self.parent_signal.current_instrument, HSTInstrument):
            # HST Health and Safety.  Only applicable to HST instruments.
            self.brightest_pixel_rate = self.parent_signal.calc_brightest_pixel(self.rate_no_qyield,self.bg_pix_rate)
            self.detector_total_rate = self.parent_signal.calc_total_detector(self.rate_no_qyield,self.bg_pix_rate)

    def signal_products(self):
        """
        The sole purpose of this function is to allow modes that use CombinedSignal to compute a reverse
        calculation.
        
        It performs the equivalent recalculations by re-running the signal_products() method on each of the
        DetectorSignals that went into this CombinedSignal, and then re-running __init__ to recreate the
        CombinedSignal from scratch.

        """
        for signal in self.signal_list:
            signal.signal_products()

        self.__init__(self.signal_list)

    def get_saturation_mask(self, rate=None):
        """
        Compute a numpy array indicating pixels with full saturation (2), partial saturation (1) and no saturation (0).
        This version just wraps whats implemented within DetectorSignal.

        Parameters
        ----------
        rate: None or 2D np.ndarray
            Detector plane rate image used to build saturation map from

        Returns
        -------
        mask: 2D np.ndarray
            Saturation mask image
        """
        if rate is None:
            rate = self.rate_plus_bg_unbinned

        mask = self.parent_signal.get_saturation_mask(rate=rate)
        return mask

    def spectral_detector_transform(self):
        """
        Create engine API format dict section containing properties of wavelength coordinates
        at the detector plane.

        Returns
        -------
        t: dict (engine API compliant keys)
        """
        return self.parent_signal.spectral_detector_transform()

    def spectral_model_transform(self):
        """
        Create engine API format dict section containing properties of the wavelength coordinates
        used in the construction of a ConvolvedSceneCube.

        Returns
        -------
        t: dict (engine API compliant keys)
        """
        return self.parent_signal.spectral_model_transform()

    def cube_wcs_info(self):
        """
        Create WCS headers and FITS binary table that describe the cube's coordinate system.

        Returns
        -------
        tbhdu: astropy.io.fits.BinTableHDU instance
            The wavelength sampling is irregularly spaced so we define a binary FITS
            table that contains the array of wavelengths, self.wave.
        header: dict
            Contains the WCS keys that define the coordinate transformation for all axes
        """
        return self.parent_signal.cube_wcs_info()


    def export_to_fits(self, fitsfile='ModelDetectorCube'):
        """
        Write convolved scene cube to a FITS file
        """
        header = self.grid.wcs_info()
        slice_index = 0
        for s,signal in enumerate(self.signal_list):
            for slice_index,flux_plus_bg in enumerate(signal.flux_plus_bg_list):
                fitsfile_slice = '{}_{}_{}.fits'.format(fitsfile, str(slice_index).strip(), s)
                t1 = fits.PrimaryHDU(np.moveaxis(flux_plus_bg, [0, 1, 2], [1, 2, 0]))
                c1 = fits.Column(name='wavelength', array=signal.wave, format='D')
                t2 = fits.BinTableHDU.from_columns([c1])
                t1.header.update(header)
                tbhdu = fits.HDUList([t1,t2])
                tbhdu.writeto(fitsfile_slice, overwrite=True)
                tbhdu.close()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import numpy as np
from astropy.io import ascii, fits
from astropy.table import unique
import astropy.units as u

from .config import DefaultConfig
from .constants import PANDEIA_WAVEUNITS, PANDEIA_FLUXUNITS
from .custom_exceptions import DataConfigurationError, EngineInputError
from .utils import get_key_list, get_dict_from_keys, recursive_subclasses, merge_data
from .pandeia_warnings import standard_warning_messages
from .io_utils import read_json
from . import config as cf

default_refdata_directory = cf.default_refdata_directory


class Extinction(DefaultConfig):

    """
    Class for handling configuration and application of spectrum NormalizationFactory

    Attributes
    ----------
    law: string
        Specifies which extinction law to use
    unit: string
        Units of extinction, either 'nh' for hydrogen column density (in cm^-2) or 'mag' for magnitudes of
        extinction in a provided bandpass, 'bandpass'
    value: float
        Amount of extinction in unit
    bandpass: string
        If unit is 'mag', this is provided to specify the bandpass over which the extinction is integrated
    extinction_wave: np.array
        Wavelength vector of the configured extinction law (microns)
    extinction_curve: np.array
        Configured extinction law in units of cross-section per hydrogen atom, cm^2/H
    """

    def __init__(self, webapp=False, config={}, **kwargs):
        """
        How an extinction law is configured depends on the defaults, the per-law defaults, and any input parameters.

        Parameters
        ----------
        webapp: bool
            Determines whether strict API checks should be performed
        config: dict
            Extra configuration information in engine input API dict format
        **kwargs: keyword/value pairs
            Extra configuration information in kwargs format
        """
        self.webapp = webapp

        # send webapp=False here since we do the API checking further below
        DefaultConfig.__init__(self, webapp=False, config=config, **kwargs)

        all_config = merge_data(config, dict(**kwargs))

        if not hasattr(self, "law"):
            self.law = self.__class__.__name__.lower()

        if self.law in self.laws:
            law_conf = self.laws[self.law]
            if "defaults" in law_conf:
                law_defaults = law_conf.pop("defaults")
            else:
                law_defaults = {}
            law_conf = merge_data(law_conf, law_defaults, config, dict(**kwargs))
            self.__dict__ = merge_data(self.__dict__, law_conf)
        else:
            msg = "Unsupported extinction law, %s" % self.law
            raise EngineInputError(value=msg)

        # do some sanity checks to make sure inputs make sense
        try:
            self._sanity_checks()
        except AttributeError as e:
            self.warnings['no_sanity_check'] = standard_warning_messages['no_sanity_check'] % (self.__class__.__name__, e)

        # add 'bandpass' to api_parameters if we need it
        if self.unit == "mag":
            self.api_parameters.append('bandpass')

        # we need to run the API checks here after merging in the normalization type-specific defaults
        if self.webapp:
            self._api_checks(all_config)

        # load the extinction curve, i.e. extinction cross section versus wavelength
        self.extinction_wave, self.extinction_curve = self._load_curve()

    def _sanity_checks(self):
        # make sure we're using a valid unit
        if self.unit not in self.units:
            msg = "Invalid extinction unit, %s, specified ." % self.unit
            raise EngineInputError(value=msg)

        # make sure if we're using a bandpass, we have it defined
        if self.unit == "mag":
            if self.bandpass not in self.bandpasses:
                msg = "Invalid extinction bandpass, %s, specified ." % self.bandpass
                raise EngineInputError(value=msg)

    def _get_config(self):
        """
        Read default configuration from JSON

        Returns
        -------
        config: dict
            All desired class attributes should have defaults defined in the config file
        """
        # get the configuration data from the pandeia reference data directory
        ref_dir = os.path.join(default_refdata_directory, "extinction")
        config = read_json(os.path.join(ref_dir, "config.json"), raise_except=True)

        config = self._load_dict(config, "required")
        # failing to load the API will result in garbage in generated calculations: they
        # use the .as_dict() method provided by DefaultConfig, which returns EVERYTHING in
        # the instance unless there's an api_parameters attribute.
        config = self._load_dict(config, "api")
        # pop the defaults entry out
        if self.webapp:
            config.pop("defaults")
        else:
            config = self._load_dict(config, "defaults")
        return config

    def _load_curve_data(self):
        """
        Use the configuration data for a given curve to configure the astropy.io.ascii fixed-width reader and load the data.

        Returns
        -------
        data: dict-like
            Data columns read from extinction data file.
        """
        curve_config = self.laws[self.law]
        filename = os.path.join(default_refdata_directory, "extinction", "curves", curve_config['filename'])
        data = ascii.read(
            filename,
            Reader=ascii.FixedWidth,
            data_start=curve_config['data_start'],
            names=curve_config['names'],
            col_starts=curve_config['col_starts'],
            col_ends=curve_config['col_ends'],
            guess=False
        )
        # make sure the data is sorted in order of increasing wavelength. 'data' is an astropy.table.Table object so has
        # built-in method to do this.
        data.sort(['wave'])

        # some of the files have duplicate rows so use astropy.table.unique() to weed them out, too
        uniq_data = unique(data, keys='wave')
        return uniq_data

    def extinction(self, wave, flux):
        """
        This is to apply the actual extinction to an input (wave, flux) set of vectors.  This is primarily
        based on comments in Issue #503#issuecomment-232136894.


        Parameters
        ----------
        wave - vector of wavelengths in ?? units
        flux - vector of flux (same shape as wave vector)

        Returns
        -------
        wave - a copy of the input vector of wavelengths
        flux - extincted version of the flux.

        """
        # interpolate the extinction curve onto the input wavelength set
        extinction_curve_interp = np.interp(wave, self.extinction_wave, self.extinction_curve)

        if self.unit == "nh":
            # if self.unit is 'nh', then self.value is a hydrogen column density in cm^-2 and we use it to scale the extinction
            # curve, C_ext, directly.
            nh_cext = self.value * extinction_curve_interp / PANDEIA_FLUXUNITS
        elif self.unit == "mag":
            # the other possible unit is 'mag' in which case we use the provided bandpass to
            # normalize the extinction curve. self._sanity_checks will raise exception before we reach
            # this point if self.unit is not 'nh' or 'mag'.
            keys = get_key_list(self.bandpass, separator=',')  # should be [photsys, filter]
            bp_filename = get_dict_from_keys(self.bandpasses, keys)['filename']

            bp_path = os.path.join(os.environ.get("PYSYN_CDBS"), "comp", "nonhst", bp_filename)

            bp_fits = fits.open(bp_path)
            bp_wave, bp_throughput = [], []
            for r in bp_fits[1].data:
                bp_wave.append(r[0])
                bp_throughput.append(r[1])
            bp_fits.close()
            bp_wave, bp_throughput = np.array(bp_wave) << u.AA, np.array(bp_throughput) << u.mJy

            # Need to interpolate the bandpass / throughput bp.wave and bp.throughput onto the self.wave
            # values and then we can do the normalization.
            # np.interp is piecewise linear
            bp_throughput_interp = np.interp(self.extinction_wave, bp_wave, bp_throughput)

            # Calculate the bandpass normalized extinction curve value
            #
            # There's more work to be done here. The normalization formally also requires the SED of the source as well.
            # This simplification is good enough for now and matches what's most commonly used.  However, for complicated
            # SEDs subject to large amounts of complicated extinction, the more formal approach will be required.
            # The work required for this is described in Issue #1884.
            bandpass_normalized_c_ext = np.sum(np.multiply(bp_throughput_interp, self.extinction_curve)) / \
                                        np.sum(bp_throughput_interp)

            # Calculate A_lambda
            # see Issue #503#issuecomment-232136894
            A_lambda = self.value * extinction_curve_interp / bandpass_normalized_c_ext
            # convert from A_lambda in magnitudes to NH * C_ext
            nh_cext = A_lambda / 1.08574
        else:
            msg = "Unsupported extinction unit, %s" % self.unit
            raise EngineInputError(value=msg)

        # Apply the extinction to the input flux
        extincted_flux = np.exp(-nh_cext) * flux

        return wave, extincted_flux


class WD2001(Extinction):

    """
    This class implements the extinction models of Weingartner & Draine 2001, ApJ, 548, 296 with updates where applicable
    from Draine 2003a, b, c.  The extinction laws from these models are tabulated and cover wavelengths from 1 A to 1 cm
    (1e-4 to 1e4 microns).  These models calculate the interstellar extinction using a dust model of carbonaceous grains
    and amorphous silicate grains.  The carbonaceous grains are like PAHs when small and like graphite when large.  The
    model has been calculated for different grain size distributions to match the observed extinction along a variety of
    lines of sight.  The extinction laws calculated from this model include:
        - Milky Way for an R_V value of 3.1 (default)
        - Milky Way for an R_V value of 4.0
        - Milky Way for an R_V value of 5.5
        - High-latitude molecular cloud HD 210121 with C/H = b_C = 40 ppm in log-normal size dists
        - Average for the LMC with C/H = b_C = 20 ppm in log-normal size dists
        - LMC with C/H = b_C = 10 ppm in log-normal size dists
        - SMC bar with C/H = b_C = 0 ppm in log-normal size dists
    """

    def _load_curve(self):
        """
        The WD2001 data files provide the extinction cross section directly in units of cm^2/H. No further
        conversion is needed except for checking of the wavelength order.

        Returns
        -------
        wave: np.array
            Wavelength vector of extinction curve
        c_ext: np.array
            Extinction curve
        """
        data = self._load_curve_data()

        wave = np.asarray(data['wave']) << PANDEIA_WAVEUNITS
        c_ext = np.asarray(data['c_ext']) << PANDEIA_FLUXUNITS

        return wave, c_ext


class Chapman09(Extinction):

    """
    This class implements the extinction law model of Chapman et al. 2009, ApJ, 690, 496.  It is most appropriate for dense
    molecular clouds and has been derived using mid-IR observations of a set of three molecular clouds: Ophiuchus, Perseus,
    and Serpens.
    """

    def _load_curve(self):
        """
        The Chapman et al. (2009) model provides the absorption and scattering cross sections separately. These need to be
        combined and then converted from cm^2/g to cm^2/H.

        Returns
        -------
        wave: np.array
            Wavelength vector of extinction curve
        c_ext: np.array
            Extinction curve
        """
        data = self._load_curve_data()

        # get the wave, C_absorption, and C_scattering data out
        wave = np.asarray(data['wave']) << PANDEIA_WAVEUNITS
        ca = np.asarray(data['ca'])
        cs = np.asarray(data['cs'])

        # add ca and cs to get total cross section and convert cm^2/g to cm^2/H.  use the WD01 R_V=5.5 factor of 2.2e-26 g/H
        # times the dust/ice mass fraction of 1.5 => 3.3e-26 g/H
        c_ext = (3.3e-26 * (ca + cs)) << PANDEIA_FLUXUNITS

        return wave, c_ext

class MWRV31(WD2001):
    """
    This class wraps the md_rv_31 extinction law that is derived from the Weingartner & Draine model for Milky Way
    extinction with R_V = 3.1
    """
    pass


class MWRV40(WD2001):
    """
    This class wraps the md_rv_40 extinction law that is derived from the Weingartner & Draine model for Milky Way
    extinction with R_V = 4.0
    """
    pass


class MWRV55(WD2001):
    """
    This class wraps the md_rv_55 extinction law that is derived from the Weingartner & Draine model for Milky Way
    extinction with R_V = 5.5
    """
    pass


class HD210121(WD2001):
    """
    This class wraps the hd210121 extinction law that is derived from the Weingartner & Draine model for extinction along
    the line of sight to the high latitude molecular cloud HD 210121.
    """
    pass


class LMCavg(WD2001):
    """
    This class wraps the lmc_avg extinction law that is derived from the Weingartner & Draine model for average extinction
    within the LMC.
    """
    pass


class LMC2(WD2001):
    """
    This class wraps the lmc_2 extinction law that is derived from the Weingartner & Draine model for a modified versions
    of the LMC extinction model with C/H = b_C = 10 ppm in log-normal size distributions.
    """
    pass


class SMCbar(WD2001):
    """
    This class wraps the smc_bar extinction law that is derived from the Weingartner & Draine model for extinction within
    the bar region of the SMC.
    """
    pass


def ExtinctionFactory(webapp=False, config={}, **kwargs):

    """
    Function to take extinction configuration data and build/return a configured instance of the appropriate
    subclass of Extinction.

    Parameters
    ----------
    webapp: bool
        Switch to toggle strict API checking
    config: dict
        Configuration data in engine API dict format
    **kwargs: keyword/value pairs
        Additional configuration data
    """
    all_config = merge_data(config, dict(**kwargs))
    if 'law' in all_config:
        method = all_config['law'].replace('_', '')
    else:
        msg = "Must specify extinction law to use."
        raise EngineInputError(value=msg)

    laws = recursive_subclasses(Extinction)
    methods = [l.__name__.lower() for l in laws]
    law_map = dict(list(zip(methods, laws)))

    if method not in methods:
        msg = "Unsupported or not yet implemented extinction law: %s" % method
        raise EngineInputError(value=msg)
    else:
        cls = law_map[method](webapp=webapp, config=config, **kwargs)
        return cls



import urllib
import json

from pandeia.engine.helpers.background.jwst.accessor_globals import bmg_url
from pandeia.engine.custom_exceptions import BMGError


def get_in_field_bg(ra, dec, date, ra_dec_str, date_str, components=11):

    '''
    component is a bitmap of modes:

        zodi = 01
        ism = 02
        cib = 08

        combined = 11
    '''
    data = {
        'query_type': 'in_field',
        'ra_dec': [ra, dec],
        'date': date,
        'components': components,
        'request_id': '1'
    }

    data_json = json.dumps(data)
    request = '%s/bmg/bmgws/?bmg=%s' % (bmg_url, data_json)
    request = request.replace(' ', '')

    print()
    print('BMG REQUEST:')
    print(request)
    print()

    try:
        response = urllib.request.urlopen(request, timeout=10)
    except urllib.error.HTTPError as e:
        # handle HTTP errors
        http_status = str(e.code)
        raise BMGError(http_status, ra_dec_str, date_str)

    response = response.read()
    response = json.loads(response)

    status = response['status']
    if status != 0:
        raise BMGError('Non-zero status code returned from BMG for request: '+request, ra_dec_str, date_str)
    else:
        [wave, bmg_bg] = response['spectrum']
        return wave, bmg_bg


def get_zodi(ra, dec, date, ra_dec_str, date_str):
    """
    Wrap get_in_field_bg() to only return the zodiacal background component by using "components=1"
    """
    return get_in_field_bg(ra, dec, date, ra_dec_str, date_str, components=1)


def get_ism(ra, dec, date, ra_dec_str, date_str):
    """
    Wrap get_in_field_bg() to only return the ISM (aka cirrus) background component by using "components=2"
    """
    return get_in_field_bg(ra, dec, date, ra_dec_str, date_str, components=2)


def get_cib(ra, dec, date, ra_dec_str, date_str):
    """
    Wrap get_in_field_bg() to only return the cosmic infrared background component by using "components=8"
    """
    return get_in_field_bg(ra, dec, date, ra_dec_str, date_str, components=8)
